// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.
//
// 
#include "stdafx.h"
#include <math.h>

#include "GLtoPS.h"

using namespace std;

// the gouraudtriangle PostScript fragement below is free.
// It is written by Frederic Delhoume (delhoume@ilog.fr).
static char *gouraudtriangleEPS[] =
{
  "/bd{bind def}bind def /triangle { aload pop   setrgbcolor  aload pop 5 3",
  "roll 4 2 roll 3 2 roll exch moveto lineto lineto closepath fill } bd",
  "/computediff1 { 2 copy sub abs threshold ge {pop pop pop true} { exch 2",
  "index sub abs threshold ge { pop pop true} { sub abs threshold ge } ifelse",
  "} ifelse } bd /computediff3 { 3 copy 0 get 3 1 roll 0 get 3 1 roll 0 get",
  "computediff1 {true} { 3 copy 1 get 3 1 roll 1 get 3 1 roll 1 get",
  "computediff1 {true} { 3 copy 2 get 3 1 roll  2 get 3 1 roll 2 get",
  "computediff1 } ifelse } ifelse } bd /middlecolor { aload pop 4 -1 roll",
  "aload pop 4 -1 roll add 2 div 5 1 roll 3 -1 roll add 2 div 3 1 roll add 2",
  "div 3 1 roll exch 3 array astore } bd /gouraudtriangle { computediff3 { 4",
  "-1 roll aload 7 1 roll 6 -1 roll pop 3 -1 roll pop add 2 div 3 1 roll add",
  "2 div exch 3 -1 roll aload 7 1 roll exch pop 4 -1 roll pop add 2 div 3 1",
  "roll add 2 div exch 3 -1 roll aload 7 1 roll pop 3 -1 roll pop add 2 div 3",
  "1 roll add 2 div exch 7 3 roll 10 -3 roll dup 3 index middlecolor 4 1 roll",
  "2 copy middlecolor 4 1 roll 3 copy pop middlecolor 4 1 roll 13 -1 roll",
  "aload pop 17 index 6 index 15 index 19 index 6 index 17 index 6 array",
  "astore 10 index 10 index 14 index gouraudtriangle 17 index 5 index 17",
  "index 19 index 5 index 19 index 6 array astore 10 index 9 index 13 index",
  "gouraudtriangle 13 index 16 index 5 index 15 index 18 index 5 index 6",
  "array astore 12 index 12 index 9 index gouraudtriangle 17 index 16 index",
  "15 index 19 index 18 index 17 index 6 array astore 10 index 12 index 14",
  "index gouraudtriangle 18 {pop} repeat } { aload pop 5 3 roll aload pop 7 3",
  "roll aload pop 9 3 roll 4 index 6 index 4 index add add 3 div 10 1 roll 7",
  "index 5 index 3 index add add 3 div 10 1 roll 6 index 4 index 2 index add",
  "add 3 div 10 1 roll 9 {pop} repeat 3 array astore triangle } ifelse } bd",
  NULL
};

GLtoPS::GLtoPS(const string &m_name, int m_buff_size)
{
	gourand = 0.075f;

	smooth_line = 0.06f;
	lineWidth = 1.0f;
	pointSize = 1.0f;
	occulsion_grid = 200;

	grid_granularity = 500;
	num_grid_x = 10;
	num_grid_y = 10;

	fout.open(m_name.c_str());

	if(!fout){
		throw "GLtoPS: Invalid file handle";
	}

	if(m_buff_size <= 0){
		// Default buffer size = (1024)^2
		feedback_size = 1048576;
	}
	else{
		feedback_size = m_buff_size;
	}

	feedback = new GLfloat [feedback_size];
	
	if(feedback == NULL){
		throw "GLtoPS: Unable to allocate memory for feedback buffer";
	}
}

GLtoPS::~GLtoPS()
{
	// Clean up the feedback buffer
	if(feedback){
		delete [] feedback;
	}

	// Close the file
	if(fout){
		fout.close();
	}
}

void GLtoPS::Start()
{
	glFeedbackBuffer(feedback_size, GL_3D_COLOR, feedback);
	glRenderMode(GL_FEEDBACK);
}

bool GLtoPS::Stop()
{

	int num_obj = glRenderMode(GL_RENDER);
	
	// Was the buffer big enough to render all of the OpenGL objects?
	if(num_obj < 0){
		
		// Double the buffer size before trying again
		feedback_size *= 2;
		
		if(feedback){
			delete [] feedback;
		}

		feedback = new GLfloat [feedback_size];
		
		if(feedback == NULL)
			throw "GLtoPS: Unable to allocate memory for feedback buffer";

		return false;
	}

	WritePS(num_obj);

	return true;
}

void GLtoPS::WritePS(int m_num_obj)
{
	int i;

	// Read back a bunch of OpenGL state to help make the EPS
    // consistent with the OpenGL clear color, line width, point
    // size, and viewport.
	glGetFloatv(GL_VIEWPORT, viewport);
	glGetFloatv(GL_COLOR_CLEAR_VALUE, clearColor);
	glGetFloatv(GL_LINE_WIDTH, &lineWidth);
	glGetFloatv(GL_POINT_SIZE, &pointSize);

	/* Emit EPS header. */
	fout << "%!PS-Adobe-2.0 EPSF-2.0" << endl;
	/* Notice %% for a single % in the fprintf calls. */
	fout << "%%Creator: Genomorama (using OpenGL feedback)" << endl;
	fout << "%%BoundingBox: " << viewport[0] << " "
		 << viewport[1] << " "
		 << viewport[2] << " "
		 << viewport[3] << endl;
	fout << "%EndComments" << endl;
	fout << endl;
	fout << "gsave" << endl;
	fout << endl;

	// Output Frederic Delhoume's "gouraudtriangle" PostScript
    // fragment.
	fout << "% the gouraudtriangle PostScript fragement below is free" << endl;
	fout << "% written by Frederic Delhoume (delhoume@ilog.fr)" << endl;

	fout << "/threshold " << gourand << " def" << endl;
	for (i = 0; gouraudtriangleEPS[i]; i++) {
		fout << gouraudtriangleEPS[i] << endl;
	}

	fout << "\n" << lineWidth <<" setlinewidth" << endl;

	/* Clear the background like OpenGL had it. */
	fout << clearColor[0] << " "
		 << clearColor[1] << " "  
		 << clearColor[2] 
		 << " setrgbcolor" << endl;

	// Note that Adobe Illustrator 8.0 doesn't like the rectfill command.
	//fprintf(file, "%g %g %g %g rectfill\n\n",
	// viewport[0], viewport[1], viewport[2], viewport[3]);
	// The following block of PostScript is equivalent to the rectfill command:
	fout << "gsave" << endl;
	fout << "newpath" << endl;
	fout << viewport[0] << " " << viewport[1] << " moveto" << endl;
	fout << viewport[2] << " 0 rlineto" << endl;
	fout << "0 " << viewport[3] << " rlineto" << endl;
	fout << viewport[2] << " neg 0 rlineto" << endl;
	fout << "closepath" << endl;
	fout << "fill" << endl;
	fout << "grestore" << endl << endl;

	WriteFeedBack(m_num_obj);

	/* Emit EPS trailer. */
	fout << "grestore" << endl << endl;
	fout << "%Add `showpage' to the end of this file to be able to print to a printer." << endl;
}

void GLtoPS::WriteFeedBack(int m_num_obj)
{
	int token;
	GLfloat *loc, *end;
	Feedback3Dcolor *vertex;
	GLfloat depthSum;
	int nprimitives, item;
	DepthIndex *prims;
	int nvertices, i;

	end = feedback + m_num_obj;

	/* Count how many primitives there are. */
	nprimitives = 0;
	loc = feedback;
	while (loc < end) {
		token = (int)*loc;
		loc++;
		switch (token) {
			case GL_LINE_TOKEN:
			case GL_LINE_RESET_TOKEN:
				loc += 14;
				nprimitives++;
				break;
			case GL_POLYGON_TOKEN:
				nvertices = (int)*loc;
				loc++;
				loc += (7 * nvertices);
				nprimitives++;
				break;
			case GL_POINT_TOKEN:
				loc += 7;
				nprimitives++;
				break;
			case GL_BITMAP_TOKEN:
				loc += 7;
				nprimitives++;
				break;
			default:
				throw "GLtoPS: Unexpected feedback token";
		}
	}

	// Allocate an array of pointers that will point back at
    // primitives in the feedback buffer.  There will be one
    // entry per primitive.  This array is also where we keep the
    // primitive's average depth.  There is one entry per
    // primitive  in the feedback buffer.
	prims = new DepthIndex [nprimitives];

	if(prims == NULL){
		throw "GLtoPS: Unable to allocate memory for PS primitives";
	}

	item = 0;
	loc = feedback;
	while (loc < end) {
		token = (int)*loc;
		prims[item].ptr = loc; // Save this primitive's location.
		loc++;
		switch (token) {
			case GL_LINE_TOKEN:
			case GL_LINE_RESET_TOKEN:
				vertex = (Feedback3Dcolor *) loc;
				depthSum = vertex[0].z + vertex[1].z;
				prims[item].depth = depthSum / 2.0f;
				loc += 14;
				break;
			case GL_POLYGON_TOKEN:
				nvertices = (int)*loc;
				loc++;
				vertex = (Feedback3Dcolor *) loc;
				depthSum = vertex[0].z;
				for (i = 1; i < nvertices; i++) {
					depthSum += vertex[i].z;
				}
				prims[item].depth = depthSum / nvertices;
				loc += (7 * nvertices);
				break;
			case GL_POINT_TOKEN:
				vertex = (Feedback3Dcolor *) loc;
				prims[item].depth = vertex[0].z;
				loc += 7;
				break;
			case GL_BITMAP_TOKEN:
				vertex = (Feedback3Dcolor *) loc;
				prims[item].depth = vertex[0].z;
				loc += 7;
				break;
			default:
				throw "GLtoPS: Unknown OpenGL feed back token!";
		}
		item++;
	}

	// Emit the Encapsulated PostScript for the primitives in
    // back to front order.
	for (item = 0; item < nprimitives; item++) {
		WritePrimitive(prims[item].ptr);
	}

  delete [] prims;
}

void GLtoPS::WritePrimitive(GLfloat * loc)
{
	int token;
	int nvertices, i;
	GLfloat red, green, blue;
	int smooth;
	GLfloat dx, dy, dr, dg, db, absR, absG, absB, colormax;
	int steps;
	Feedback3Dcolor *vertex;
	GLfloat xstep = 0.0f, 
			ystep = 0.0f, 
			rstep = 0.0f, 
			gstep = 0.0f, 
			bstep = 0.0f;
	GLfloat xnext = 0.0f, 
			ynext = 0.0f, 
			rnext = 0.0f, 
			gnext = 0.0f, 
			bnext = 0.0f, 
			distance = 0.0f;

	token = (int)*loc;
	loc++;
	switch (token) {
		case GL_LINE_RESET_TOKEN:
		case GL_LINE_TOKEN:
			vertex = (Feedback3Dcolor *) loc;

			dr = vertex[1].red - vertex[0].red;
			dg = vertex[1].green - vertex[0].green;
			db = vertex[1].blue - vertex[0].blue;

			if (dr != 0 || dg != 0 || db != 0) {
				/* Smooth shaded line. */
				dx = vertex[1].x - vertex[0].x;
				dy = vertex[1].y - vertex[0].y;
				distance = (float)sqrt(dx * dx + dy * dy);

				absR = (float)fabs(dr);
				absG = (float)fabs(dg);
				absB = (float)fabs(db);

				colormax = max(absR, max(absG, absB));
				steps = (int)max(1.0f, colormax * distance * smooth_line);

				xstep = dx / steps;
				ystep = dy / steps;

				rstep = dr / steps;
				gstep = dg / steps;
				bstep = db / steps;

				xnext = vertex[0].x;
				ynext = vertex[0].y;
				rnext = vertex[0].red;
				gnext = vertex[0].green;
				bnext = vertex[0].blue;

				// Back up half a step; we want the end points to be
				// exactly the their endpoint colors.
				xnext -= xstep / 2.0f;
				ynext -= ystep / 2.0f;
				rnext -= rstep / 2.0f;
				gnext -= gstep / 2.0f;
				bnext -= bstep / 2.0f;
			} else {
				/* Single color line. */
				steps = 0;
			}

			fout << vertex[0].red << " "
				 << vertex[0].green << " "
				 << vertex[0].blue
				 << " setrgbcolor" << endl;

			fout << vertex[0].x << " " << vertex[0].y << " moveto" << endl;

			for (i = 0; i < steps; i++) {
				xnext += xstep;
				ynext += ystep;
				rnext += rstep;
				gnext += gstep;
				bnext += bstep;
				fout << xnext << " " 
					 << ynext << " lineto stroke" << endl;
				fout << rnext << " " 
					 << gnext << " "
					 << bnext << " setrgbcolor" << endl;

				fout << xnext << " " << ynext << " moveto" << endl;
			}

			fout << vertex[1].x << " " << vertex[1].y << " lineto stroke" << endl;

			// Each vertex element in the feedback
			// buffer is 7 GLfloats.
			// loc += 14;          

			break;
		case GL_POLYGON_TOKEN:
			nvertices = (int)*loc;
			loc++;
			vertex = (Feedback3Dcolor *) loc;

			if (nvertices > 0) {
				red = vertex[0].red;
				green = vertex[0].green;
				blue = vertex[0].blue;
				smooth = 0;
				for (i = 1; i < nvertices; i++) {
					if (red != vertex[i].red || green != vertex[i].green || blue != vertex[i].blue) {
						smooth = 1;
						break;
					}
				}
				if (smooth) {
					/* Smooth shaded polygon; varying colors at vetices. */
					int triOffset;

					/* Break polygon into "nvertices-2" triangle fans. */
					for (i = 0; i < nvertices - 2; i++) {
						triOffset = i * 7;
						fout << "["
							 << vertex[0].x << " "
							 << vertex[i + 1].x << " "
							 << vertex[i + 2].x << " "
							 << vertex[0].y << " " 
							 << vertex[i + 1].y << " " 
							 << vertex[i + 2].y << "]";

						fout << " ["
							<< vertex[0].red << " "
							<< vertex[0].green << " "
							<< vertex[0].blue << "] ["
							<< vertex[i + 1].red << " "
							<< vertex[i + 1].green << " "
							<< vertex[i + 1].blue << "] ["
							<< vertex[i + 2].red << " "
							<< vertex[i + 2].green << " "
							<< vertex[i + 2].blue 
							<< "] gouraudtriangle" << endl;
					}
				} else {
					/* Flat shaded polygon; all vertex colors the same. */
					fout << "newpath" << endl;
					fout << red << " "
						 << green << " "
						 << blue << " setrgbcolor" << endl;

					/* Draw a filled triangle. */
					fout << vertex[0].x << " "
						 << vertex[0].y << " moveto" << endl;

					for (i = 1; i < nvertices; i++) {
						fout << vertex[i].x << " "
							 << vertex[i].y << " lineto" << endl;
					}

					fout << "closepath fill" << endl << endl;
				}
			}

			break;
		case GL_POINT_TOKEN:
			vertex = (Feedback3Dcolor *) loc;

			fout << vertex[0].red << " "
				 << vertex[0].green << " "
				 << vertex[0].blue << " setrgbcolor" << endl;

			fout << vertex[0].x << " "
				 << vertex[0].y << " "
				 << pointSize*0.5f << " 0 360 arc fill" << endl << endl;
			break;
		case GL_BITMAP_TOKEN:
			// Do nothing for now. Eventually will need to use the pass through functionality
			// to describe the bitmaps for rendering (like storing the CallList id).
			break;
		default:
		  throw "GLtoPS: Unexpected feedback token";
	}
}
