// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.
//
//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#ifndef __GL_TO_PS
#define	__GL_TO_PS

// Include the OpenGL headers
#ifdef _MAC_OSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include <string>
#include <fstream>

/* OpenGL's GL_3D_COLOR feedback vertex format. */
typedef struct _Feedback3Dcolor {
	GLfloat x;
	GLfloat y;
	GLfloat z;
	GLfloat red;
	GLfloat green;
	GLfloat blue;
	GLfloat alpha;
} Feedback3Dcolor;

typedef struct _DepthIndex {
	GLfloat *ptr;
	GLfloat depth;
	GLbyte visible;
} DepthIndex;

class GLtoPS{
private:
	GLfloat *feedback;
	GLint feedback_size;
	std::ofstream fout;

	GLfloat clearColor[4], viewport[4];
	GLfloat lineWidth, pointSize;
	GLint occulsion_grid;
	GLfloat gourand;
	GLfloat smooth_line;

	GLint grid_granularity;
	GLint num_grid_x;
	GLint num_grid_y;

	void WritePS(int m_num_obj);

	void WriteFeedBack(int m_num_obj);

	void WritePrimitive(GLfloat * loc);

public:
	GLtoPS(const std::string &m_name, int m_buff_size = -1);
	~GLtoPS();

	void Start();
	bool Stop();
};

#endif // __GL_TO_PS

