#include <OpenGL/glu.h>
#include <OpenGL/glext.h>

#include "font_file.h"
#include "ncbi_util.h"
#include "univout.h"
#include "primer.h"

#include "glCheck.h"
#include "main.h"
#include <sstream>

// Local funcions
pascal void LiveActionProc( ControlHandle control, SInt16 part );
pascal void LiveActionHScrollProc( ControlHandle control, SInt16 part );
pascal void LiveActionVScrollProc( ControlHandle control, SInt16 part );

enum
{
    kMainMenu = 500,
    kCloseMenuItem = 2,
    kInfoMenuItem = 4,
    kAnimateMenuItem = 5,
    kInfoState = 1,
    kAnimateState = 1
};
// The main window widgets
const ControlID kHorizontalScrollBar = { 'MAIN', 1 };
const ControlID kVerticalScrollBar = { 'MAIN', 2 };

// Enumerate the find dialog widgets here
const ControlID kFindSearchKey          = { 'FIND', 1 };
const ControlID kFindSearchString       = { 'FIND', 2 };
const ControlID kFindSearchResults      = { 'FIND', 3 };
const ControlID kFindSearchResultsView  = { 'FIND', 4 };

// Enumerate the hybridize dialog widgets here
const ControlID kHybridStrandEdit			= { 'HYBD', 1 };
const ControlID kHybridSaltEdit				= { 'HYBD', 2 };
const ControlID kHybridTmEdit				= { 'HYBD', 3 };
const ControlID kHybridPrimer1Edit          = { 'HYBD', 4 };
const ControlID kHybridPrimer2Edit          = { 'HYBD', 5 };
const ControlID kHybridExactEdit			= { 'HYBD', 6 };
const ControlID kHybridProbeEdit			= { 'HYBD', 7 };
const ControlID kHybridResultsEdit          = { 'HYBD', 8 };
const ControlID kHybridResultsView          = { 'HYBD', 9 };
const ControlID kHybridMaxAmpliconEdit		= { 'HYBD', 10 };
const ControlID kHybridMinAmpliconEdit		= { 'HYBD', 11 };

// Enumerate the MOLigo dialog widgets here
const ControlID kMoligoStrandEdit			= { 'MOLI', 1 };
const ControlID kMoligoSaltEdit				= { 'MOLI', 2 };
const ControlID kMoligoTmEdit				= { 'MOLI', 3 };
const ControlID kMoligoPrimer1Edit          = { 'MOLI', 4 };
const ControlID kMoligoPrimer2Edit          = { 'MOLI', 5 };
const ControlID kMoligoExactEdit			= { 'MOLI', 6 };
const ControlID kMoligoResultsEdit          = { 'MOLI', 7 };
const ControlID kMoligoResultsView          = { 'MOLI', 8 };

// Enumerate the primer Prediction dialog widgets here
const ControlID kPrimerLengthEdit			= { 'PRMR', 1 };
const ControlID kPrimerTmEdit				= { 'PRMR', 2 };
const ControlID kPrimerHairpinTmEdit		= { 'PRMR', 3 };
const ControlID kPrimerSaltEdit				= { 'PRMR', 4 };
const ControlID kPrimerStrandEdit			= { 'PRMR', 5 };
const ControlID kPrimerPoly3GCCheck			= { 'PRMR', 6 };
const ControlID kPrimerMulti5GCCheck		= { 'PRMR', 7 };
const ControlID kPrimerRunCheck				= { 'PRMR', 8 };
const ControlID kPrimerRunLengthEdit		= { 'PRMR', 9 };
const ControlID kPrimerGCCheck				= { 'PRMR', 10 };
const ControlID kPrimerMinGCEdit			= { 'PRMR', 11 };
const ControlID kPrimerMaxGCEdit			= { 'PRMR', 12 };
const ControlID kPrimerTCheck				= { 'PRMR', 13 };

// Enumerate the text info dialog widgets here
const ControlID kInfoText          = { 'INFO', 1 };
const ControlID kInfoView          = { 'INFO', 2 };

const ControlID kDownloadSearchKey		= { 'DOWN', 1};
const ControlID kDownloadSearchString   = { 'DOWN', 2};
const ControlID kDownloadSearchResults  = { 'DOWN', 3};

const ControlID kOptionsFont			= { 'OPTN', 1 };
const ControlID kOptionsGC				= { 'OPTN', 2 };
const ControlID kOptionsTm				= { 'OPTN', 3 };
const ControlID kOptionsTmSalt			= { 'OPTN', 4 };
const ControlID kOptionsTmStrand		= { 'OPTN', 5 };

const ControlID kOptionsFontValue		= { 'OPTN', 6 };
const ControlID kOptionsGCValue			= { 'OPTN', 7 };
const ControlID kOptionsTmValue			= { 'OPTN', 8 };
const ControlID kOptionsTmSaltValue		= { 'OPTN', 9 };
const ControlID kOptionsTmStrandValue   = { 'OPTN', 10 };

// Enumerate the export dialog widgets here
const ControlID kExportFormat			= { 'EXPT', 0 };
const ControlID kExportAnnotGene		= { 'EXPT', 1 };
const ControlID kExportAnnotPseudo		= { 'EXPT', 2 };
const ControlID kExportAnnotCDS			= { 'EXPT', 3 };
const ControlID kExportAnnotRNA			= { 'EXPT', 4 };
const ControlID kExportAnnottRNA		= { 'EXPT', 5 };
const ControlID kExportAnnotMisc		= { 'EXPT', 6 };
const ControlID kExportAnnotprimer		= { 'EXPT', 7 };
const ControlID kExportAnnotUser		= { 'EXPT', 8 };
const ControlID kExportAnnotIntergenic  = { 'EXPT', 9 };

// Enumerate the edit dialog widgets here
const ControlID kAnnotRange			= { 'ANOT', 1 };
const ControlID kAnnotType			= { 'ANOT', 2 };
const ControlID kAnnotStrand		= { 'ANOT', 3 };
const ControlID kAnnotAccession		= { 'ANOT', 4 };
const ControlID kAnnotGI			= { 'ANOT', 5 };
const ControlID kAnnotEC			= { 'ANOT', 6 };
const ControlID kAnnotLocus			= { 'ANOT', 7 };
const ControlID kAnnotLocusTag		= { 'ANOT', 8 };
const ControlID kAnnotProduct		= { 'ANOT', 9 };
const ControlID kAnnotNote			= { 'ANOT', 10 };
const ControlID kAnnotNoteView		= { 'ANOT', 11 };

EventHandlerUPP gEvtHandler;			// main event handler
EventHandlerUPP gWinEvtHandler;			// window event handler

// The find window
WindowRef find_window = NULL;

// The primer window
WindowRef primer_window = NULL;

// The hybridization window
WindowRef hybrid_window = NULL;

// The MOLigo window
WindowRef moligo_window = NULL;

// The info window
WindowRef info_window = NULL;

// The download window
WindowRef download_window = NULL;

// The options window
WindowRef option_window = NULL;

// The export window
WindowRef export_window = NULL;

// The annotation edit window
WindowRef annotation_window = NULL;

ControlActionUPP fProc = NewControlActionUPP(LiveActionProc);
ControlActionUPP hscrollProc = NewControlActionUPP(LiveActionHScrollProc);
ControlActionUPP vscrollProc = NewControlActionUPP(LiveActionVScrollProc);
	
// Store Download search results
std::list< std::pair<unsigned int, std::string> > docsum_list;

// Store the state of the annotation edit dialog
selected_annotation_info edit_annot;

IBNibRef nibRef = NULL;

AbsoluteTime gStartTime;

char gErrorMessage[256] = ""; // buffer for error message output
float gErrorTime = 0.0;

#pragma mark ---- OpenGL Capabilities ----

// GL configuration info globals
// see glcheck.h for more info
GLCaps * gDisplayCaps = NULL; // array of GLCaps
CGDisplayCount gNumDisplays = 0;
// related DM change notification:
DMExtendedNotificationUPP gConfigEDMUPP = NULL;

// The default local of the horizontal slider
#define DEFAULT_HORIZONTAL_SCROLL_BAR_LOC 50

// The default local of the vertical slider
#define DEFAULT_VERTICAL_SCROLL_BAR_LOC 50

static void getCurrentCaps (void)
{
 	// Check for existing opengl caps here
	// This can be called again with same display caps array when display configurations are changed and
	//   your info needs to be updated.  Note, if you are doing dynmaic allocation, the number of displays
	//   may change and thus you should always reallocate your display caps array.
	if (gDisplayCaps && HaveOpenGLCapsChanged (gDisplayCaps, gNumDisplays)) { // see if caps have changed
		free (gDisplayCaps);
		gDisplayCaps = NULL;
	}
	if (!gDisplayCaps) { // if we do not have caps
		CheckOpenGLCaps (0, NULL, &gNumDisplays); // will just update number of displays
		gDisplayCaps = (GLCaps*) malloc (sizeof (GLCaps) * gNumDisplays);
		CheckOpenGLCaps (gNumDisplays, gDisplayCaps, &gNumDisplays);
	}
}

#pragma mark ---- Utilities ----

// Add a scrollable text widget to a dialog
OSStatus AddTextViewWidget(WindowRef m_window, const ControlID &m_view_id, const ControlID &m_text_id,
	const bool &m_read_only)
{
    OSStatus result = noErr;
    HIRect boundsRect;
    HIViewRef scrollView = NULL;
    HIViewRef textView = NULL;
    HIViewRef contentView = NULL;
    HIViewRef dialogView = NULL;
    
    // Get the placeholder view
    HIViewFindByID(HIViewGetRoot(m_window), m_view_id, &contentView);
    
    if(contentView == NULL){
        return controlHandleInvalidErr;
    }
    
    // Get the dialog view
    HIViewFindByID(HIViewGetRoot(m_window), kHIViewWindowContentID, &dialogView);
    
    if(contentView == NULL){
        return controlHandleInvalidErr;
    }
    
    // The bounds of the placeholder view ...
    HIViewGetBounds(contentView, &boundsRect);

    // ... in the coordinates of the dialog
    HIViewConvertRect(&boundsRect, contentView, dialogView);
    
    result = HIScrollViewCreate(kHIScrollViewValidOptions, &scrollView);
    
    if(result != noErr){
        return result; 
    }
    
    HIViewSetFrame(scrollView, &boundsRect);
    
    HIViewSetVisible(scrollView, true);
	
    // Add to the dialog, not the placeholder
    result = HIViewAddSubview(dialogView, scrollView);
    
    if(result != noErr){
        return result;
    }
	
	if(m_read_only){
		result = HITextViewCreate(NULL, 0, 
			kTXNReadOnlyMask | kTXNWantHScrollBarMask | kTXNWantVScrollBarMask, 
			&textView);
	}
	else{
		result = HITextViewCreate(NULL, 0, 
			kTXNWantHScrollBarMask | kTXNWantVScrollBarMask, 
			&textView);
	}
	
    if(result != noErr){
        return result;
    }
    
    TXNMargins margins;
    
    margins.topMargin = 0;
    margins.leftMargin = 0;
    margins.bottomMargin = 10;
    margins.rightMargin = 10;
                           
    const TXNControlTag state_tag[] = {kTXNWordWrapStateTag, kTXNMarginsTag};
    const TXNControlData state_data[] = {kTXNNoAutoWrap, (long unsigned int)(&margins)};
    
    // Disable automatic line wrapping
    TXNSetTXNObjectControls(HITextViewGetTXNObject(textView), 
        false, 2, state_tag, state_data);
	
    HIViewSetVisible(textView, true);
    
    result = HIViewAddSubview(scrollView, textView);
    
    if(result != noErr){
        return result;
    }

	HILayoutInfo layoutInfo;

	// Bind the view to its parent on all sides (affects resize)
	layoutInfo.binding.left.kind = kHILayoutBindLeft;
	layoutInfo.binding.left.toView = NULL;
	layoutInfo.binding.left.offset = 0.0;
	layoutInfo.binding.top.kind = kHILayoutBindTop;
	layoutInfo.binding.top.toView = NULL;
	layoutInfo.binding.top.offset = 0.0;
	layoutInfo.binding.right.kind = kHILayoutBindRight;
	layoutInfo.binding.right.toView = NULL;
	layoutInfo.binding.right.offset = 0.0;
	layoutInfo.binding.bottom.kind = kHILayoutBindBottom;
	layoutInfo.binding.bottom.toView = NULL;
	layoutInfo.binding.bottom.offset = 0.0;
	
	HIViewSetLayoutInfo(scrollView, &layoutInfo);
	
    SetControlID(textView, &m_text_id);
    
    return noErr;
}

static pascal OSStatus DownloadItemDataCallback(ControlRef browser, 
    DataBrowserItemID itemID, DataBrowserPropertyID property, 
    DataBrowserItemDataRef itemData, Boolean changeValue)
{
    CFStringRef         text;
    std::stringstream   sout;
    OSStatus            err = noErr;
    std::list< std::pair<unsigned int, std::string> >::iterator iter;
    unsigned int count = 1;
    
    for(iter = docsum_list.begin();(iter != docsum_list.end()) && (count != itemID);iter++, count++){
        // Do nothing!
    }
    
    if(iter == docsum_list.end()){
        // There is not data to display!
        return err;
    }
    
    switch(property){
        case 'desc': // Description
            // Format a discription string to populate the Databrowser
            sout << "(" << iter->first << ") " << iter->second;
            
            text = CFStringCreateWithCString(NULL, sout.str().c_str(), CFStringGetSystemEncoding());
			
            err = SetDataBrowserItemDataText(itemData, text); 
            
            CFRelease(text);
            break;
        case kDataBrowserItemIsSelectableProperty:
            err = SetDataBrowserItemDataBooleanValue(itemData, true);
            break;
        case kDataBrowserItemIsActiveProperty:
            err = SetDataBrowserItemDataBooleanValue(itemData, true);
            break;	
        case kDataBrowserItemIsEditableProperty:
            err = SetDataBrowserItemDataBooleanValue(itemData, false);
            break;
        default:
            err = errDataBrowserPropertyNotSupported;
            break;
    };
    
    return err;
}

OSStatus ConfigureDataBrowser(WindowRef m_window, const ControlID &m_id)
{
    ControlRef browser;
    OSStatus result = noErr;
    
    DataBrowserCallbacks myCallbacks;
    DataBrowserViewStyle viewStyle;
    
    // Get a reference to the control
    result = GetControlByID(m_window, &m_id, &browser);
     
    if(result != noErr){
        return result;
    }
    
    // Use latest layout and callback signatures
    myCallbacks.version = kDataBrowserLatestCallbacks;
    
    result = InitDataBrowserCallbacks(&myCallbacks);
    
    if(result != noErr){
        return result;
    }
    
    myCallbacks.u.v1.itemDataCallback = 
        NewDataBrowserItemDataUPP(DownloadItemDataCallback);
    
    result = SetDataBrowserCallbacks(browser, &myCallbacks);

    if(result != noErr){
        return result;
    }
    
    // After adding callbacks, configure the browser
    GetDataBrowserViewStyle(browser, &viewStyle);
    
    
    if(viewStyle != kDataBrowserListView){
        throw __FILE__ ":ConfigureDataBrowser: Please use a list view databrowser"; 
    }
    
    return result;
}

//-----------------------------------------------------------------------------
//	 SetStaticTextText
//-----------------------------------------------------------------------------
//	Sets the text of an edit text control and optionally redraws it.
//
OSStatus
SetStaticTextText( ControlHandle control, ConstStr255Param text, Boolean draw )
{
	OSStatus	err;
	
	if ( control == nil )
		return paramErr;

	err = SetControlData( control, 0, kControlStaticTextTextTag, text[0], (Ptr)(text+1) );
	if ( (err == noErr) && draw )
		DrawOneControl( control );
	
	return err;
}

void UpdateOptionsDialog(pRecContext pContextInfo)
{
	ControlRef control_ref;
	Str255 valueText;
	char buffer[256];
	
	if((pContextInfo == NULL) || (pContextInfo->ptrGP == NULL)){
		return;
	}
	
	// Set the slider 
	GetControlByID(option_window, &kOptionsFont, &control_ref);
	SetControlValue(control_ref, pContextInfo->ptrGP->font_scale_index());

	// Set the dialog value
	GetControlByID(option_window, &kOptionsFontValue, &control_ref );
	sprintf(buffer, "%f", pContextInfo->ptrGP->font_scale());
	CopyCStringToPascal(buffer, valueText);
	SetStaticTextText(control_ref, valueText, true);
	
	
	// Set the slider 
	GetControlByID(option_window, &kOptionsGC, &control_ref);
	SetControlValue(control_ref, pContextInfo->ptrGP->content_window());

	// Set the dialog value
	GetControlByID(option_window, &kOptionsGCValue, &control_ref );
	sprintf(buffer, "%d", pContextInfo->ptrGP->content_window());
	CopyCStringToPascal(buffer, valueText);
	SetStaticTextText(control_ref, valueText, true);

	// Set the slider 
	GetControlByID(option_window, &kOptionsTm, &control_ref);
	SetControlValue(control_ref, pContextInfo->ptrGP->tm_window());

	// Set the dialog value
	GetControlByID(option_window, &kOptionsTmValue, &control_ref );
	sprintf(buffer, "%d", pContextInfo->ptrGP->tm_window());
	CopyCStringToPascal(buffer, valueText);
	SetStaticTextText(control_ref, valueText, true);

	// Set the slider 
	GetControlByID(option_window, &kOptionsTmSalt, &control_ref);
	SetControlValue(control_ref, pContextInfo->ptrGP->tm_salt_index());

	// Set the dialog value
	GetControlByID(option_window, &kOptionsTmSaltValue, &control_ref );
	sprintf(buffer, "%.3g", pContextInfo->ptrGP->tm_salt());
	CopyCStringToPascal(buffer, valueText);
	SetStaticTextText(control_ref, valueText, true);

	// Set the slider 
	GetControlByID(option_window, &kOptionsTmStrand, &control_ref);
	SetControlValue(control_ref, pContextInfo->ptrGP->tm_strand_index());

	// Set the dialog value
	GetControlByID(option_window, &kOptionsTmStrandValue, &control_ref );
	sprintf(buffer, "%.3g", pContextInfo->ptrGP->tm_strand());
	CopyCStringToPascal(buffer, valueText);
	SetStaticTextText(control_ref, valueText, true);
}

pascal void LiveActionHScrollProc(ControlHandle control, SInt16 part)
{
	WindowRef window = FrontNonFloatingWindow();
	pRecContext pContextInfo = GetCurrentContextInfo(window);
	
	if( (pContextInfo == NULL) || (pContextInfo->ptrGP == NULL) ){
		return;
	}
	
	SInt16 currentValue = GetControl32BitValue(control);
    SInt16 min_scroll = GetControl32BitMinimum(control);
    SInt16 max_scroll = GetControl32BitMaximum(control);
	
	const SInt16 delta_page = 10;
	const SInt16 delta_arrow = 1;
	
	switch (part) {

        case kControlUpButtonPart:
			pContextInfo->ptrGP->scroll_bar(currentValue - delta_arrow, max_scroll - min_scroll);
			pContextInfo->ptrGP->scroll_bar(DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			
			break;
			
        case kControlDownButtonPart:
			pContextInfo->ptrGP->scroll_bar(currentValue + delta_arrow, max_scroll - min_scroll);
			pContextInfo->ptrGP->scroll_bar(DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			
            break;

        case kControlPageUpPart:
			pContextInfo->ptrGP->scroll_bar(currentValue - delta_page, max_scroll - min_scroll);
			pContextInfo->ptrGP->scroll_bar(DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			
            break;

        case kControlPageDownPart:
			pContextInfo->ptrGP->scroll_bar(currentValue + delta_page, max_scroll - min_scroll);
			pContextInfo->ptrGP->scroll_bar(DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
			
			break;
			
		default: // The slider
			pContextInfo->ptrGP->scroll_bar(currentValue, max_scroll - min_scroll);
			break;
    };
	
	// Update the display "manually". Note that InvalWindowRect() is
	// blocked unil the user releases the mouse (hence the direct call to
	// drawGL()).
	drawGL(window, true);
}

pascal void LiveActionVScrollProc(ControlHandle control, SInt16 part)
{
	WindowRef window = FrontNonFloatingWindow();
	pRecContext pContextInfo = GetCurrentContextInfo(window);
	
	if( (pContextInfo == NULL) || (pContextInfo->ptrGP == NULL) ){
		return;
	}
	
	SInt16 currentValue = GetControl32BitValue(control);
    SInt16 min_scroll = GetControl32BitMinimum(control);
    SInt16 max_scroll = GetControl32BitMaximum(control);
	
	const SInt16 delta_page = 10;
	const SInt16 delta_arrow = 1;
	
	switch (part) {

        case kControlUpButtonPart:
			pContextInfo->ptrGP->zoom_bar(currentValue - delta_arrow, max_scroll - min_scroll);
			pContextInfo->ptrGP->zoom_bar(DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			
			break;
			
        case kControlDownButtonPart:
			pContextInfo->ptrGP->zoom_bar(currentValue + delta_arrow, max_scroll - min_scroll);
			pContextInfo->ptrGP->zoom_bar(DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			
            break;

        case kControlPageUpPart:
			pContextInfo->ptrGP->zoom_bar(currentValue - delta_page, max_scroll - min_scroll);
			pContextInfo->ptrGP->zoom_bar(DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			
            break;

        case kControlPageDownPart:
			pContextInfo->ptrGP->zoom_bar(currentValue + delta_page, max_scroll - min_scroll);
			pContextInfo->ptrGP->zoom_bar(DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			SetControl32BitValue(control, DEFAULT_VERTICAL_SCROLL_BAR_LOC);
			
			break;
			
		default: // The slider
			pContextInfo->ptrGP->zoom_bar(currentValue, max_scroll - min_scroll);
			break;
    };
	
	// Update the display "manually". Note that InvalWindowRect() is
	// blocked unil the user releases the mouse (hence the direct call to
	// drawGL()).
	drawGL(window, true);
}

pascal void LiveActionProc( ControlHandle control, SInt16 part )
{
	ControlRef control_ref;
	WindowRef window = FrontNonFloatingWindow();
	pRecContext pContextInfo = GetCurrentContextInfo(window);
	Str255 valueText;
	char buffer[256];

	SInt16 slide_value = GetControlValue(control);
	
	if((window == NULL) || (pContextInfo == NULL) || (pContextInfo->ptrGP == NULL)){
		return;
	}
	
	GetControlByID(option_window, &kOptionsFont, &control_ref);
	
	if(control_ref == control){
		
		// Update the genomes
		pContextInfo->ptrGP->font_scale_index(slide_value);
		
		GetControlByID(option_window, &kOptionsFontValue, &control_ref );
		
		// Set the dialog value
		sprintf(buffer, "%f", pContextInfo->ptrGP->font_scale());
		
		CopyCStringToPascal(buffer, valueText);
		
		SetStaticTextText(control_ref, valueText, true);
		
		// Update the display "manually". Note that InvalWindowRect() is
		// blocked unil the user releases the mouse (hence the direct call to
		// drawGL()).
		drawGL(window, true);
		
		return;
	}
	
	GetControlByID( option_window, &kOptionsGC, &control_ref );
	
	if(control_ref == control){
		
		// Update the genomes
		pContextInfo->ptrGP->content_window(slide_value);
		
		GetControlByID(option_window, &kOptionsGCValue, &control_ref );
		
		// Set the dialog value
		sprintf(buffer, "%d", pContextInfo->ptrGP->content_window());
		
		CopyCStringToPascal(buffer, valueText);
		
		SetStaticTextText(control_ref, valueText, true);
		
		// Update the display "manually". Note that InvalWindowRect() is
		// blocked unil the user releases the mouse (hence the direct call to
		// drawGL()).
		drawGL(window, true);
		
		return;
	}

	GetControlByID( option_window, &kOptionsTm, &control_ref );
	
	if(control_ref == control){
		
		// Update the genomes
		pContextInfo->ptrGP->tm_window(slide_value);
		
		GetControlByID(option_window, &kOptionsTmValue, &control_ref );
		
		// Set the dialog value
		sprintf(buffer, "%d", pContextInfo->ptrGP->tm_window());
		
		CopyCStringToPascal(buffer, valueText);
		
		SetStaticTextText(control_ref, valueText, true);
		
		// Update the display "manually". Note that InvalWindowRect() is
		// blocked unil the user releases the mouse (hence the direct call to
		// drawGL()).
		drawGL(window, true);

		return;
	}


	GetControlByID( option_window, &kOptionsTmSalt, &control_ref );
	
	if(control_ref == control){
		
		// Update the genomes
		pContextInfo->ptrGP->tm_salt_index(slide_value);
		
		GetControlByID(option_window, &kOptionsTmSaltValue, &control_ref );
		
		// Set the dialog value
		sprintf(buffer, "%.3g", pContextInfo->ptrGP->tm_salt());
		
		CopyCStringToPascal(buffer, valueText);
		
		SetStaticTextText(control_ref, valueText, true);
		
		// Update the display "manually". Note that InvalWindowRect() is
		// blocked unil the user releases the mouse (hence the direct call to
		// drawGL()).
		drawGL(window, true);

		return;
	}


	GetControlByID( option_window, &kOptionsTmStrand, &control_ref );
	
	if(control_ref == control){
		
		// Update the genomes
		pContextInfo->ptrGP->tm_strand_index(slide_value);
		
		GetControlByID(option_window, &kOptionsTmStrandValue, &control_ref );
		
		// Set the dialog value
		sprintf(buffer, "%.3g", pContextInfo->ptrGP->tm_strand());
		
		CopyCStringToPascal(buffer, valueText);
		
		SetStaticTextText(control_ref, valueText, true);
		
		// Update the display "manually". Note that InvalWindowRect() is
		// blocked unil the user releases the mouse (hence the direct call to
		// drawGL()).
		drawGL(window, true);

		return;
	}
}

#pragma mark ---- Error Reporting ----

// C string to Pascal string
static void cstr2pstr (StringPtr outString, const char *inString)
{	
	unsigned short x = 0;
	do {
	    outString [x + 1] = (unsigned char) inString [x];
		x++;
	} while ((inString [x] != 0)  && (x < 256));
	outString [0] = x;									
}

OSStatus ErrorDialog(const char *primary_error, const char *secondary_error = NULL)
{
	// Display an alert dialog box to let the user know what happened
	SInt16 outItemHit;
	Str255 error_string_pascal;
	OSStatus result = noErr;
	
	// Format the string for use in a Carbon Alert dialog box
	cstr2pstr (error_string_pascal, primary_error);
	
	if(secondary_error != NULL){
		Str255 secondary_error_string_pascal;
		cstr2pstr (secondary_error_string_pascal, secondary_error);
		
		result = StandardAlert (
		kAlertCautionAlert,
		error_string_pascal,
		secondary_error_string_pascal,
		NULL,
		&outItemHit); // Which button did the user press?
	}
	else{
		result = StandardAlert (
			kAlertCautionAlert,
			error_string_pascal,
			NULL,
			NULL,
			&outItemHit); // Which button did the user press?
	}
	
	return result;
}

// error reporting as both window message and debugger string
void reportError (char * strError)
{
	ErrorDialog(strError);
}

// Open a dialog box to recive user filenames to open
bool GetMultiFiles(std::list<std::string> &m_filenames)
{
    NavDialogCreationOptions inOptions;
    NavTypeListHandle inTypeList;
    NavDialogRef outDialog;
    NavReplyRecord outReply;
    OSStatus result = noErr;
    NavUserAction action;
    
    // Set the dialog options
    result = NavGetDefaultDialogCreationOptions(&inOptions);

    if(noErr != result){
        throw __FILE__ ":GetFiles: Error in NavGetDefaultDialogCreationOptions";
    }
    
    inTypeList = (NavTypeListHandle)GetResource('open', 128);
    
    result = NavCreateGetFileDialog(&inOptions, 
        inTypeList, 
        NULL, 
        NULL,
        NULL,
        NULL,
        &outDialog);

    if(noErr != result){
        throw __FILE__ ":GetFiles: Error in NavCreateGetFileDialog";
    }
    
    result = NavDialogRun(outDialog);
    
    if(noErr != result){
        throw __FILE__ ":GetFiles: Error in NavDialogRun";
    }
    
    action = NavDialogGetUserAction(outDialog);

    // Do we need to process file name(s)?    
    if((action != kNavUserActionCancel) && (action != kNavUserActionNone)){
    
        result = NavDialogGetReply(outDialog, &outReply);
    
        if(noErr != result){
            throw __FILE__ ":GetFiles: Error in NavDialogGetReply";
        }
        
        // Is this data structure valid?
        if(outReply.validRecord){
            SInt32 file_count;
            
            result = AECountItems(&outReply.selection, &file_count);
            
            if(result != noErr){
                throw __FILE__ ":GetFiles: Error in AECountItems";
            }
            
			try{
				for(int i = 1;i <= file_count;i++){
					AEKeyword key_word;
					AEDesc desc_result;
					
					// Request the output in FSRef format instead of FSSpec format.
					// This will allow the use of the FSRefMakePath function to extract the
					// filename as a plain old vanilla C-style string.
					result = AEGetNthDesc (
						&outReply.selection,
						i,
						typeFSRef,
						&key_word,
						&desc_result);
						
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEGetNthDesc";
					}
					
					// Get the file name.
					FSRef theRef;
					
					result = AEGetDescData(&desc_result, &theRef, sizeof(FSRef));
					
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEGetDescData";
					}
					
					// Convert the FSRef to a path name
					char filename[512];
					
					FSRefMakePath(&theRef, (UInt8*)filename, 512);
					
					// Save the filename
					m_filenames.push_back(filename);
					
					result = AEDisposeDesc(&desc_result);
					
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEDisposeDesc";
					}
				}
            }
			catch(const char *error){
				
				ErrorDialog(error);
			}
			catch(...){
				
				ErrorDialog("Unhandled error opening file", 
					"Please report this error to jgans@lanl.gov");
			}
			
            result = NavDisposeReply(&outReply);
    
            if(noErr != result){
                throw __FILE__ ":OpenFiles: Error in NavDisposeReply";
            }

        }
    }
    
    NavDialogDispose(outDialog);
    
    if (inTypeList != NULL){
        ReleaseResource( (char**)inTypeList );
    }
    
    return true;
}

bool GetSaveFileName(std::string &m_filename)
{
    NavDialogCreationOptions inOptions;
    NavDialogRef outDialog;
    NavReplyRecord outReply;
    OSStatus result = noErr;
    NavUserAction action;
	
    OSType fileTypeToSave = 'TEXT';
    OSType creatorType = 'xAPP';
    
	bool status = false;
	
	// Set the dialog options	
    result = NavGetDefaultDialogCreationOptions(&inOptions);

    if(noErr != result){
        throw __FILE__ ":GetSaveFileName: Error in NavGetDefaultDialogCreationOptions";
    }
	
    result = NavCreatePutFileDialog(&inOptions,
        fileTypeToSave,
        creatorType,
        NULL,
        NULL,
        &outDialog);
    
    if(noErr != result){
        throw __FILE__ ":GetSaveFileName: Error in NavCreateChooseFileDialog";
    }
    
    result = NavDialogRun(outDialog);
    
    if(noErr != result){
        throw __FILE__ ":GetSaveFileName: Error in NavDialogRun";
    }
    
    action = NavDialogGetUserAction(outDialog);

    // Do we need to process file name(s)?    
    if((action != kNavUserActionCancel) && (action != kNavUserActionNone)){
    
        result = NavDialogGetReply(outDialog, &outReply);
    
        if(noErr != result){
            throw __FILE__ ":GetSaveFileName: Error in NavDialogGetReply";
        }
        
        // Is this data structure valid?
        if(outReply.validRecord){
            
            // Note that we can't JUST use the saveFileName field, because it doesn't contain
            // the complete path to the file!
            AEKeyword key_word;
            AEDesc desc_result;
            
            // Request the output in FSRef format instead of FSSpec format.
            // This will allow the use of the FSRefMakePath function to extract the
            // filename as a plain old vanilla C-style string.
            result = AEGetNthDesc (
                &outReply.selection,
                1, // Get the first (and only) descriptor -- 1 based!
                typeFSRef,
                &key_word,
                &desc_result);
                
            if(result != noErr){
                throw __FILE__ ":GetSaveFileName: Error in AEGetNthDesc";
            }
            
            // Get the file name.
            FSRef theRef;
                
            result = AEGetDescData(&desc_result, &theRef, sizeof(FSRef));
            
            if(result != noErr){
                throw __FILE__ ":GetSaveFileName: Error in AEGetDescData";
            }
            
            // Convert the FSRef to a path name
            char path[512];
            
            FSRefMakePath(&theRef, (UInt8*)path, 512);
            
            // Now get the filename
            int len = CFStringGetLength(outReply.saveFileName) + 1;
            char *file = new char [len];
            
            CFStringGetCString(outReply.saveFileName, file, len, CFStringGetSystemEncoding());
            
            // Whew! Now we have a complete path!
            m_filename = std::string(path) + "/" + std::string(file);
            
            if(file){
                delete [] file;
            }
            
            result = AEDisposeDesc(&desc_result);
            
            if(result != noErr){
                throw __FILE__ ":GetSaveFileName: Error in AEDisposeDesc";
            }
        
            result = NavDisposeReply(&outReply);
    
            if(noErr != result){
                throw __FILE__ ":GetSaveFileName: Error in NavDisposeReply";
            }
			
			status = true;
        }
    }
    
    NavDialogDispose(outDialog);
        
    return status;

}

// if error dump agl errors to debugger string, return error
OSStatus aglReportError (void)
{
	GLenum err = aglGetError();
	if (AGL_NO_ERROR != err) {
		char errStr[256];
		sprintf (errStr, "AGL: %s",(char *) aglErrorString(err));
		reportError (errStr);
	}
	// ensure we are returning an OSStatus noErr if no error condition
	if (err == AGL_NO_ERROR)
		return noErr;
	else
		return (OSStatus) err;
}

// ---------------------------------

// if error dump gl errors to debugger string, return error
OSStatus glReportError (void)
{
	GLenum err = glGetError();
	if (GL_NO_ERROR != err) {
		char errStr[256];
		sprintf (errStr, "GL: %s",(char *) gluErrorString(err));
		reportError (errStr);
	}
	// ensure we are returning an OSStatus noErr if no error condition
	if (err == GL_NO_ERROR)
		return noErr;
	else
		return (OSStatus) err;
}

#pragma mark ---- OpenGL Utilities ----

// handles resizing of GL, the horizontal scroll bar and the context update
OSStatus resizeGL(pRecContext pContextInfo, const Rect &m_rectPort)
{
    OSStatus err = noErr;
	
    if(!pContextInfo || !pContextInfo->aglContext || !pContextInfo->ptrGP){
        return paramErr;
    }
    
    // re-attach drawable to ensure context is updated
    if (!aglSetCurrentContext (pContextInfo->aglContext)){
        err = aglReportError ();
    }
    
    if ((noErr == err) && !aglUpdateContext (pContextInfo->aglContext)){
        err = aglReportError ();
    }
    
	const int scroll_bar_width = 15;
	
	pContextInfo->width = m_rectPort.right - m_rectPort.left - scroll_bar_width;
    pContextInfo->height = m_rectPort.bottom - m_rectPort.top - scroll_bar_width;
	
	// Make room for the horizontal scroll bar on the bottom of the screen
	// (and room for the resize tab as well!).
	GLint bufferRect[4];
	
	bufferRect[0] = 0; // 0 = left edge
	bufferRect[1] = scroll_bar_width; // 0 = bottom edge
	bufferRect[2] = pContextInfo->width; // width of buffer rect
	bufferRect[3] = pContextInfo->height; // height of buffer rect
	
	aglSetInteger(pContextInfo->aglContext, AGL_BUFFER_RECT, bufferRect);
	aglEnable(pContextInfo->aglContext, AGL_BUFFER_RECT);
	
    pContextInfo->ptrGP->setup_view_OGL(pContextInfo->width, pContextInfo->height);
    
	// Update the scroll bar position
	ControlRef scroll_bar;
	WindowRef main_window = FrontNonFloatingWindow(); // The main window
	
	// First extract the horizontal scroll bar control
    err = GetControlByID(main_window, &kHorizontalScrollBar, &scroll_bar);
    
    if(err != noErr){
        throw __FILE__ ":ResizeGL: Unable to get horizontal scroll bar control";
    }

	MoveControl(scroll_bar, m_rectPort.left, m_rectPort.bottom - scroll_bar_width);
	SizeControl(scroll_bar, pContextInfo->width, scroll_bar_width);

	// New extract the vertical scroll bar control
    err = GetControlByID(main_window, &kVerticalScrollBar, &scroll_bar);
    
    if(err != noErr){
        throw __FILE__ ":ResizeGL: Unable to get vertical scroll bar control";
    }

	MoveControl(scroll_bar, m_rectPort.right - scroll_bar_width, m_rectPort.top);
	SizeControl(scroll_bar, scroll_bar_width, pContextInfo->height);
	
    return err;
}

// ---------------------------------

#pragma mark ---- OpenGL Minimize Handler (thanks to Dan Herman) ----

void InvertGLImage( char *imageData, size_t imageSize, size_t rowBytes )
{
	long i, j;
	char *tBuffer = (char*) malloc (rowBytes);
	
	if (NULL == tBuffer) return;
		
	// Copy by rows through temp buffer
	for (i = 0, j = imageSize - rowBytes; i < (long)(imageSize >> 1); i += rowBytes, j -= rowBytes) {
		memcpy( tBuffer, &imageData[i], rowBytes );
		memcpy( &imageData[i], &imageData[j], rowBytes );
		memcpy( &imageData[j], tBuffer, rowBytes );
	}
	
	free(tBuffer);
}

// ---------------------------------

void CompositeGLBufferIntoWindow (AGLContext ctx, Rect *bufferRect, WindowRef win)
{
	GWorldPtr pGWorld;
	QDErr err;
	// blit OpenGL content into window backing store
	// allocate buffer to hold pane image
	long width  = (bufferRect->right - bufferRect->left);
	long height  = (bufferRect->bottom - bufferRect->top);
	Rect src_rect = {0, 0, height, width};
	long rowBytes = width * 4;
	long imageSize = rowBytes * height;
	char *image = (char *) NewPtr (imageSize);
	if (!image) {
		return;		// no harm in continuing
	}
	
	// pull GL content down to our image buffer
	aglSetCurrentContext( ctx );
	glReadPixels (0, 0, width, height, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, image);

	// GL buffers are upside-down relative to QD buffers, so we need to flip it
	InvertGLImage( image, imageSize, rowBytes );

	// create a GWorld containing our image
	err = NewGWorldFromPtr (&pGWorld, k32ARGBPixelFormat, &src_rect, 0, 0, 0, image, rowBytes);
	if (err != noErr) {
		free( image );
		return;
	}
	
	SetPort( GetWindowPort (win));
	CopyBits( GetPortBitMapForCopyBits (pGWorld), GetPortBitMapForCopyBits (GetWindowPort (win)), &src_rect, bufferRect, srcCopy, 0 );

	DisposeGWorld( pGWorld );
	DisposePtr ( image );
}

#pragma mark ---- Display Manager Event Handling ----

// update our GL configuration info based on display change notification
void handleConfigDMEvent (void *userData, short theMessage, void *notifyData)
{
	if (kDMNotifyEvent == theMessage) { // post change notifications only
		getCurrentCaps ();
	}
}

// ---------------------------------

// handle display config changes meaning we need to update the GL context via the resize function and check for window dimension changes
// also note we redraw the content here as it could be lost in a display config change
void handleWindowDMEvent (void *userData, short theMessage, void *notifyData)
{
	if (kDMNotifyEvent == theMessage) { // post change notifications only
		pRecContext pContextInfo = NULL;
		WindowRef window = (WindowRef) userData;
		if (window)
			pContextInfo = GetCurrentContextInfo (window);
		if (pContextInfo) { // have a valid OpenGl window
			Rect rectPort;

			GetWindowPortBounds (window, &rectPort);			
			// resizeGL (pContextInfo, rectPort.right - rectPort.left, rectPort.bottom - rectPort.top);
			resizeGL (pContextInfo, rectPort);
			InvalWindowRect (window,  &rectPort); // force redrow
		}
	}
}

#pragma mark ---- OpenGL Window ----

// intializes context conditions
static void initialConditions (pRecContext pContextInfo)
{
    BlockZero (pContextInfo, sizeof (recContext));
    
    pContextInfo->ptrGP = new GenomePage;
    
    if(! pContextInfo->ptrGP){
        throw __FILE__ ":initialConditions: Unable to allocate memory for GenomePage";
    }
}

// ---------------------------------

OSStatus buildGL (WindowRef window)
{
	OSStatus err = noErr;
	const unsigned int num_samples = 4;
	
	GLint vanilla_attrib[] = { AGL_RGBA, AGL_DOUBLEBUFFER, AGL_DEPTH_SIZE, 16, AGL_NONE };
	
	GLint multisample_attrib[] = { AGL_RGBA, AGL_DOUBLEBUFFER, AGL_DEPTH_SIZE, 16, AGL_NO_RECOVERY,
					   AGL_SAMPLE_BUFFERS_ARB, 1, AGL_SAMPLES_ARB, num_samples, AGL_NONE };
    
    pRecContext pContextInfo = GetCurrentContextInfo(window);
    ProcessSerialNumber psn = { 0, kCurrentProcess };
    
    if (NULL == pContextInfo){
        return paramErr;
	}
	
	if (pContextInfo->aglContext){
		return noErr; // already built
	}
	
	// build context
	pContextInfo->aglContext = NULL;
	pContextInfo->aglPixFmt = aglChoosePixelFormat(NULL, 0, multisample_attrib);
	
	err = aglReportError();
		
	if (pContextInfo->aglPixFmt) {
		pContextInfo->aglContext = aglCreateContext(pContextInfo->aglPixFmt, NULL);
		err = aglReportError();		
	}

	// if multi-sample pixel format or context fails allocate a non-multi-sample one
	if( (noErr != err) || (NULL == pContextInfo->aglContext) ){ 
		// Try the vanilla configuration
		if (pContextInfo->aglPixFmt){
			aglDestroyPixelFormat (pContextInfo->aglPixFmt);
		}
		
		pContextInfo->aglPixFmt = aglChoosePixelFormat(NULL, 0, vanilla_attrib);
		
		err = aglReportError();
		
		if(pContextInfo->aglPixFmt){
			pContextInfo->aglContext = aglCreateContext (pContextInfo->aglPixFmt, NULL);
			err = aglReportError();
		}
	}

	if (pContextInfo->aglContext) {
		//short fNum;
		GLint swap = 1;
		//CGRect viewRect = {{0.0f, 0.0f}, {0.0f, 0.0f}};

        GrafPtr portSave = NULL;
        GetPort (&portSave);
        SetPort ((GrafPtr) GetWindowPort (window));

		if(!aglSetDrawable(pContextInfo->aglContext, GetWindowPort (window)))
			err = aglReportError ();
		if (!aglSetCurrentContext(pContextInfo->aglContext))
			err = aglReportError ();
		pContextInfo->currentVS = aglGetVirtualScreen (pContextInfo->aglContext); // sync renderer
		
		// ensure we know when display configs are changed
		pContextInfo->windowEDMUPP = NewDMExtendedNotificationUPP (handleWindowDMEvent); // for display change notification
		DMRegisterExtendedNotifyProc (pContextInfo->windowEDMUPP, (void *) window, NULL, &psn);

		// VBL SYNC
		if (!aglSetInteger (pContextInfo->aglContext, AGL_SWAP_INTERVAL, &swap))
			aglReportError ();
		
        // Initialize OpenGL here!
        if(pContextInfo->ptrGP){
            Rect rectPort;
            
            GetWindowPortBounds (window, &rectPort);

            pContextInfo->ptrGP->initialize_OGL();
            
            // Load the font information here
            LoadFont(pContextInfo->ptrGP->get_font_metric());
            
            // Make sure all sizes are set correctly
            //resizeGL(pContextInfo, rectPort.right - rectPort.left, rectPort.bottom - rectPort.top);
			resizeGL(pContextInfo, rectPort);
        }
        
        SetPort (portSave);
    }
    return err;
}

// ---------------------------------

// dump window data structures and OpenGL context and related data structures
OSStatus disposeGL (pRecContext pContextInfo)
{
	// release our data
    if ( pContextInfo != NULL )
    {
        if (pContextInfo->windowEDMUPP) { // dispose UPP for DM notifications
            DisposeDMExtendedNotificationUPP (pContextInfo->windowEDMUPP);
            pContextInfo->windowEDMUPP = NULL;
        }
        
        aglSetDrawable (pContextInfo->aglContext, NULL);
        aglSetCurrentContext (NULL);
        
        if (pContextInfo->aglContext) {
            aglDestroyContext (pContextInfo->aglContext);
            pContextInfo->aglContext = NULL;
        }
        
        if (pContextInfo->aglPixFmt) {
            aglDestroyPixelFormat (pContextInfo->aglPixFmt);
            pContextInfo->aglPixFmt = NULL;
        }
        
        if (pContextInfo->timer) {
            RemoveEventLoopTimer(pContextInfo->timer);
            pContextInfo->timer = NULL;
        }
		
        if(pContextInfo->ptrGP){
            delete pContextInfo->ptrGP;
            pContextInfo->ptrGP = NULL;
        }
    }
    
    return noErr;
}

// ---------------------------------

// main OpenGL drawing function
void drawGL (WindowRef window, Boolean swap)
{    
    pRecContext pContextInfo = NULL;
    
    if(!window){
        return;
    }
    
    pContextInfo = GetCurrentContextInfo (window);
    
    if(!pContextInfo->aglContext){
        buildGL(window);
    }
    
    if(!pContextInfo->ptrGP){
        return;
    }
    
    // Add all openGL drawing code here!
	
    if(pContextInfo->ptrGP->render() == true){
		Rect rectPort;
		// Get the window bounds and invalidate the window
		GetWindowPortBounds(window, &rectPort);
		InvalWindowRect(window, &rectPort);
	}
    
    if(swap){
        aglSwapBuffers (pContextInfo->aglContext);
    }
}

#pragma mark ---- Carbon Timer ----

// per-window timer function, basic time based animation preformed here
static void timerContextCB (WindowRef window)
{
    pRecContext pContextInfo = NULL;
    
    if (window){
        pContextInfo = GetCurrentContextInfo (window);
    }
    
    if (pContextInfo) {
        AbsoluteTime currTime = UpTime ();
        double deltaTime = (float) AbsoluteDeltaToDuration (currTime, pContextInfo->time);
        pContextInfo->time = currTime;	// reset for next time interval
        if (0 > deltaTime)	// if negative microseconds
                deltaTime /= -1000000.0;
        else				// else milliseconds
                deltaTime /= 1000.0;
		
        // Add timer code here!
    }
}


// timer callback bottleneck
static pascal void timerCB (EventLoopTimerRef inTimer, void* userData)
{
    #pragma unused (inTimer)
    timerContextCB ((WindowRef) userData); // timer based update
}

// ---------------------------------



// get UPP for timer
static EventLoopTimerUPP getTimerUPP (void)
{
	static EventLoopTimerUPP	sTimerUPP = NULL;
	
	if (sTimerUPP == NULL)
		sTimerUPP = NewEventLoopTimerUPP (timerCB);
	
	return sTimerUPP;
}

// ---------------------------------

WindowRef createNewWindow(void)
{
    EventHandlerRef ref;
    EventTypeSpec list[] = { { kEventClassWindow, kEventWindowCollapsing },
                             { kEventClassWindow, kEventWindowShown },
                             { kEventClassWindow, kEventWindowActivated },
                             { kEventClassWindow, kEventWindowClose },
                             { kEventClassWindow, kEventWindowDrawContent },
                             { kEventClassWindow, kEventWindowBoundsChanged },
                             { kEventClassWindow, kEventWindowZoomed },
                             { kEventClassKeyboard, kEventRawKeyDown },
                             { kEventClassKeyboard, kEventRawKeyRepeat },
                             { kEventClassKeyboard, kEventRawKeyUp } };

    WindowRef window = NULL;
	
    pRecContext pContextInfo = (pRecContext) NewPtrClear (sizeof (recContext)); // memory for window record
    initialConditions (pContextInfo);
    
    CreateWindowFromNib (nibRef, CFSTR("MainWindow"), &window); // build window
    
    if (window){
        SetWRefCon (window, (long) pContextInfo); // point to the window record in the ref con of the window
        InstallWindowEventHandler (window, gWinEvtHandler, GetEventTypeCount (list), list, (void*)window, &ref); // add event handler
        ShowWindow(window);
        
        // Disable the timer for now
        if (false && !pContextInfo->timer) {
            pContextInfo->time = UpTime ();
            InstallEventLoopTimer (GetCurrentEventLoop(), 0, 0.01, getTimerUPP (), (void *) window, &pContextInfo->timer);
        }
		
		// Add a handler for the scroll bars
		ControlRef control;
		
		GetControlByID(window, &kHorizontalScrollBar, &control);
		SetControlAction(control, hscrollProc);
		
		GetControlByID(window, &kVerticalScrollBar, &control);
		SetControlAction(control, vscrollProc);
    }
	
	return window;
}

// ---------------------------------

bool UpdateAnnotationDialog(GenomePage *ptrGP)
{
	OSStatus result = noErr;
	ControlHandle control;
	HIViewRef textView;
    TXNObject obj = NULL;
    TXNOffset start_offset, stop_offset;
	
	if(ptrGP == NULL){
        return false;
    }

	// Only update if the user has selected a single annotation
	edit_annot = ptrGP->selected_annotation();
		
	if( edit_annot.is_genome_valid() == false){
	
		ErrorDialog("Please select a genome before editing annotations");
		return false;
	}
	
	if( edit_annot.is_annot_valid() == false){
		
		// Reset all defaults
		result = StringToEditTextConrol(annotation_window, kAnnotRange, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotRange control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotLocus, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotLocus control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotLocusTag, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotLocusTag control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotProduct, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotProduct control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotEC, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotEC control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotGI, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotGI control";
		}
		
		result = StringToEditTextConrol(annotation_window, kAnnotAccession, "");
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotAccession control";
		}
		
		HIViewFindByID(HIViewGetRoot(annotation_window), kAnnotNote, &textView);
		
		obj = HITextViewGetTXNObject(textView);
		TXNSelectAll(obj);
		TXNGetSelection(obj, &start_offset, &stop_offset);
				
		TXNSetData(obj, kTXNTextData, "", 0, start_offset, stop_offset);
		
		// Display the upper left hand corner of text
		TXNSelectAll(obj);
		TXNShowSelection(obj, false);
		TXNSetSelection(obj, 0, 0);
	
		result = GetControlByID(annotation_window, &kAnnotStrand, &control);
    
		if(result != noErr){
			throw __FILE__ ":DownloadSearch: Unable to get kAnnotStrand control";
		}
		
		SetControl32BitValue(control, 1); // Seq_strand_plus is the default
		
		result = GetControlByID(annotation_window, &kAnnotType, &control);
    
		if(result != noErr){
			throw __FILE__ ":DownloadSearch: Unable to get kAnnotType control";
		}
		
		SetControl32BitValue(control, 1); // GeneAnnotation::CDS is the default
	}
	else{
	
		result = StringToEditTextConrol( annotation_window, kAnnotRange, edit_annot.range() );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotRange control";
		}
		
		result = StringToEditTextConrol( annotation_window, kAnnotLocus, 
			edit_annot.info(GeneAnnotation::LOCUS) );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotLocus control";
		}
		
		result = StringToEditTextConrol( annotation_window, kAnnotLocusTag, 
			edit_annot.info(GeneAnnotation::LOCUS_TAG) );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotLocusTag control";
		}
		
		result = StringToEditTextConrol( annotation_window, kAnnotProduct, 
			edit_annot.info(GeneAnnotation::PRODUCT) );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotProduct control";
		}
		
		result = StringToEditTextConrol( annotation_window, kAnnotEC, 
			edit_annot.info(GeneAnnotation::EC) );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotEC control";
		}
		
		if(edit_annot.gi() > 0){
			result = IntToEditTextConrol( annotation_window, kAnnotGI, 
				edit_annot.gi() );
		}
		else{
			result = StringToEditTextConrol( annotation_window, kAnnotGI, "");
		}
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotGI control";
		}
		
		result = StringToEditTextConrol( annotation_window, kAnnotAccession, 
			edit_annot.accession(true) );
		
		if(result != noErr){
			throw __FILE__ ":UpdateAnnotationDialog: Unable to get kAnnotAccession control";
		}
				
		HIViewFindByID(HIViewGetRoot(annotation_window), kAnnotNote, &textView);
		
		obj = HITextViewGetTXNObject(textView);
		TXNSelectAll(obj);
		TXNGetSelection(obj, &start_offset, &stop_offset);
		
		// *Cludge* to get the window to show scroll bars for long lines that don't
		// fill the text window
		const std::string buffer = wrap_string(edit_annot.info(GeneAnnotation::NOTE), 40);
		
		TXNSetData(obj, kTXNTextData, buffer.c_str(), buffer.size(), start_offset, stop_offset);
		
		// Display the upper left hand corner of text
		TXNSelectAll(obj);
		TXNShowSelection(obj, false);
		TXNSetSelection(obj, 0, 0);

		result = GetControlByID(annotation_window, &kAnnotStrand, &control);
    
		if(result != noErr){
			throw __FILE__ ":DownloadSearch: Unable to get kAnnotStrand control";
		}
		
		switch( edit_annot.strand() ){
			case Seq_strand_plus:
				SetControl32BitValue(control, 1);
				break;
			case Seq_strand_minus:
				SetControl32BitValue(control, 2);
				break;	
		};
		
		result = GetControlByID(annotation_window, &kAnnotType, &control);
    
		if(result != noErr){
			throw __FILE__ ":DownloadSearch: Unable to get kAnnotType control";
		}
		
		SetControl32BitValue(control, edit_annot.type() + 1);
	}	
	
	return true;
}

bool ApplyAnnotation(GenomePage *ptrGP)
{
	OSStatus result = noErr;
	ControlHandle control;
	
	if(ptrGP == NULL){
        return false;
    }
	
	try{
		
		edit_annot.range( trim_space( EditTextConrolToString(annotation_window, kAnnotRange) ) );
		
		edit_annot.info(GeneAnnotation::LOCUS, 
			trim_space( EditTextConrolToString(annotation_window, kAnnotLocus) ) );
		
		edit_annot.info(GeneAnnotation::LOCUS_TAG, 
			trim_space( EditTextConrolToString(annotation_window, kAnnotLocusTag) ) );
		
		edit_annot.info(GeneAnnotation::PRODUCT, 
			trim_space( EditTextConrolToString(annotation_window, kAnnotProduct) ) );
		
		edit_annot.info(GeneAnnotation::EC, 
			trim_space( EditTextConrolToString(annotation_window, kAnnotEC) ) );
		
		// Get the type key
		GetControlByID(annotation_window, &kAnnotType, &control);
		
		if(result != noErr){
			throw __FILE__ ":ApplyAnnotation: Unable to get annotation type control";
		}
		
		switch( GetControl32BitValue(control) ){
			case 1:
				edit_annot.type(GeneAnnotation::CDS);
				break;
			case 2: 
				edit_annot.type(GeneAnnotation::GENE);
				break;
			case 3:
				edit_annot.type(GeneAnnotation::PSEUDO_GENE);
				break;
			case 4: 
				edit_annot.type(GeneAnnotation::RNA);
				break;
			case 5: 
				edit_annot.type(GeneAnnotation::tRNA);
				break;
			case 6: 
				edit_annot.type(GeneAnnotation::IMP);
				break;
			case 7:
				edit_annot.type(GeneAnnotation::PRIMER);
				break;
			case 8:
				edit_annot.type(GeneAnnotation::USER);
				break;
			case 9:
				edit_annot.type(GeneAnnotation::NONE);
				break;

			default:
				throw __FILE__ ":ApplyAnnotation: Unknown annotation type";
		};
		
		HIViewRef textView;
		TXNObject obj = NULL;
		TXNOffset start_offset, stop_offset;
		
		HIViewFindByID(HIViewGetRoot(annotation_window), kAnnotNote, &textView);
		
		obj = HITextViewGetTXNObject(textView);
		TXNSelectAll(obj);
		TXNGetSelection(obj, &start_offset, &stop_offset);
		TXNSetSelection(obj, 0, 0);
		
		char** text_buffer = NULL;
		
		TXNGetDataEncoded(obj, start_offset, stop_offset, &text_buffer, kTXNTextData);
		
		// Use the pack_string function to remove all end of line and carrige return characters
		edit_annot.info(GeneAnnotation::NOTE, 
			pack_string( std::string(*text_buffer, stop_offset - start_offset) ) );
		
		// The documentation claims that we need to free the associated memory ...
		DisposeHandle(text_buffer);
		
		const std::string gi_str = trim_space( EditTextConrolToString(annotation_window, kAnnotGI) );
		const std::string acc_str = trim_space( EditTextConrolToString(annotation_window, kAnnotAccession) );
		
		int gi = -1;
		
		if( !gi_str.empty() ){
			
			if(integer(gi_str) == false){
				throw "gi must be a positive  integer";
			}
			
			gi = atoi( gi_str.c_str() );
			
			if(gi < 1){
				throw "gi must be a positive  integer";
			}
		}
		
		SeqIdPtr sip = NULL;
		
		if(gi > 0){
		
			sip = ValNodeNew(NULL);
			sip->choice = SEQID_GI;
			sip->data.intvalue = gi;
		}
		
		if(acc_str != ""){
			
			if(sip == NULL){
			
				if(acc_str.find('|') != std::string::npos){
					// Yes
					sip = SeqIdParse ( (char*)acc_str.c_str() );
				}
				else{
					// No
					sip = SeqIdParse( (char*)std::string("gb|" + acc_str).c_str() );
				}
				
				if(sip == NULL){
					throw "Unable to write accession";
				}
			}
			else{
				if(acc_str.find('|') != std::string::npos){
					// Yes
					sip->next = SeqIdParse ( (char*)acc_str.c_str() );
				}
				else{
					// No
					sip->next = SeqIdParse( (char*)std::string("gb|" + acc_str).c_str() );
				}
				
				if(sip->next == NULL){
					throw "Unable to write accession";
				}
			}
		}
		
		if(sip){
		
			// Only update the seqid if the user has provided the info
			edit_annot.seqid(sip);
			sip = SeqIdSetFree(sip);
		}
		
		ptrGP->update_annotation(edit_annot);
	}
	catch(const char *error){
	
		ErrorDialog(error);
		return false;
	}
	catch(...){
		ErrorDialog("Caught an unhandled error in ApplyAnnotation");
		return false;
	}
	
	return true;
}

void DisplayInfo(const std::string &m_info)
{
    HIViewRef textView;
	TXNObject obj = NULL;
	
    if(!info_window){
        return;
    }

    HIViewFindByID(HIViewGetRoot(info_window), kInfoText, &textView);
    
    // *Cludge* to get the window to show scroll bars for long lines that don't
    // fill the text window
    std::string buffer = m_info + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";
	
	obj = HITextViewGetTXNObject(textView);
	
	TXNSetData(obj, kTXNTextData, buffer.c_str(), buffer.size(), 
        kTXNStartOffset, kTXNEndOffset);
 
    ShowWindow(info_window);
	
	// Display the upper left hand corner of text
    TXNSelectAll(obj);
    TXNShowSelection(obj, false);
    TXNSetSelection(obj, 0, 0);

}

bool FindSeq(GenomePage *ptrGP)
{
    OSStatus result = noErr;
    ControlHandle   tempControl;
    HIViewRef textView;
    TXNObject obj = NULL;
    TXNOffset start_offset, stop_offset;
    Size buffer_size = -1;
    char *buffer = NULL;
    int search_key = -1;
    
    if(ptrGP == NULL){
        return false;
    }
    
    // First extract the Search key value
    result = GetControlByID(find_window, &kFindSearchKey, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":FindSeq: Unable to get search key control";
    }
    
    switch(GetControl32BitValue(tempControl)){
        case 1: // Gene Name/Tag
            search_key = GenomePage::SEARCH_KEY_NAME;
            break;
        case 2: // GI
            search_key = GenomePage::SEARCH_KEY_GI;
            break;
        case 3: // Base Number
            search_key = GenomePage::SEARCH_KEY_BASE;
            break;
        case 4: // Annotation Keyword(s)
            search_key = GenomePage::SEARCH_KEY_TEXT;
            break;
        case 5: // DNA Sequence
            search_key = GenomePage::SEARCH_KEY_DNA;
            break;
		case 6: // Genome Name
            search_key = GenomePage::SEARCH_KEY_GENOME_NAME;
            break;
        default:
            throw __FILE__ ":FindSeq: Unknown Search key value";
    };
    
    // Now Extract the search key string
    result = GetControlByID(find_window, &kFindSearchString, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":FindSeq: Unable to get search string control";
    }
    
    // Get the size of the buffer
    result = GetControlDataSize(tempControl, kControlEntireControl, 
		kControlEditTextTextTag, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":FindSeq: Unable to get search string size";
    }

    buffer = new char[buffer_size + 1];
    
    if(buffer == NULL){
        throw __FILE__ ":FindSeq: Unable to allocate buffer!";
    }
    
    result = GetControlData(tempControl, kControlEntireControl, 
		kControlEditTextTextTag, buffer_size, buffer, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":FindSeq: Unable to get search string data";
    }
    
    // Add a string terminator
    buffer[buffer_size] = '\0';
    
    std::string search_results;
    
    try{
		SetThemeCursor(kThemeWatchCursor);
        search_results = ptrGP->search(search_key, buffer);
    }
    catch(const char *error){
        search_results = error;
    }
    
	SetThemeCursor(kThemeArrowCursor);
	
    // Clean up before displaying the results
    if(buffer){
        delete [] buffer;
    }
    
    // Show the search results
    HIViewFindByID(HIViewGetRoot(find_window), kFindSearchResults, &textView);
    
    obj = HITextViewGetTXNObject(textView);
    
    TXNSelectAll(obj);
    TXNGetSelection(obj, &start_offset, &stop_offset);
    
    // *Cludge* to get the window to show scroll bars for long lines that don't
    // fill the text window
    search_results = search_results + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";

    TXNSetData(obj, kTXNTextData, search_results.c_str(), search_results.size(), start_offset, stop_offset);
    
    // Display the upper left hand corner of text
    TXNSelectAll(obj);
    TXNShowSelection(obj, false);
    TXNSetSelection(obj, 0, 0);
    
    return true;
}

bool CheckboxConrolToBool(WindowRef &m_window, const ControlID &kControl)
{
	OSStatus result = noErr;
    ControlHandle   tempControl;

	if(NULL == m_window){
		throw __FILE__ ":CheckboxConrolToBool: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":CheckboxConrolToBool: Unable to get control";
    }
    
	return (GetControl32BitValue(tempControl) != 0);
}

float EditTextConrolToFloat(WindowRef &m_window, const ControlID &kControl)
{
	OSStatus result = noErr;
    ControlHandle   tempControl;
    Size buffer_size = -1;
    char *buffer = NULL;
	
	if(NULL == m_window){
		throw __FILE__ ":EditTextConrolToFloat: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get control";
    }
    
    // Get the size of the buffer
    result = GetControlDataSize(tempControl, kControlEntireControl, kControlEditTextTextTag, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get buffer size";
    }

    buffer = new char[buffer_size + 1];
    
    if(buffer == NULL){
        throw __FILE__ ":EditTextConrolToFloat: Unable to allocate buffer!";
    }
    
    result = GetControlData(tempControl, kControlEntireControl, kControlEditTextTextTag, buffer_size, buffer, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get string data";
    }
    
    // Add a string terminator
    buffer[buffer_size] = '\0';

	float ret = atof(buffer);
	
	delete [] buffer;
	
	return ret;
}

int EditTextConrolToInt(WindowRef &m_window, const ControlID &kControl)
{
	OSStatus result = noErr;
    ControlHandle   tempControl;
    Size buffer_size = -1;
    char *buffer = NULL;
	
	if(NULL == m_window){
		throw __FILE__ ":EditTextConrolToFloat: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get control";
    }
    
    // Get the size of the buffer
    result = GetControlDataSize(tempControl, kControlEntireControl, kControlEditTextTextTag, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get buffer size";
    }

    buffer = new char[buffer_size + 1];
    
    if(buffer == NULL){
        throw __FILE__ ":EditTextConrolToFloat: Unable to allocate buffer!";
    }
    
    result = GetControlData(tempControl, kControlEntireControl, kControlEditTextTextTag, buffer_size, buffer, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToFloat: Unable to get string data";
    }
    
    // Add a string terminator
    buffer[buffer_size] = '\0';

	int ret = atoi(buffer);
	
	delete [] buffer;
	
	return ret;
}


std::string EditTextConrolToString(WindowRef &m_window, const ControlID &kControl)
{
	OSStatus result = noErr;
    ControlHandle   tempControl;
    Size buffer_size = -1;
    char *buffer = NULL;
	
	if(NULL == m_window){
		throw __FILE__ ":EditTextConrolToString: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToString: Unable to get control";
    }
    
    // Get the size of the buffer
    result = GetControlDataSize(tempControl, kControlEntireControl, kControlEditTextTextTag, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToString: Unable to get buffer size";
    }

    buffer = new char[buffer_size + 1];
    
    if(buffer == NULL){
        throw __FILE__ ":EditTextConrolToString: Unable to allocate buffer!";
    }
    
    result = GetControlData(tempControl, kControlEntireControl, kControlEditTextTextTag, buffer_size, buffer, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":EditTextConrolToString: Unable to get string data";
    }
    
    // Add a string terminator
    buffer[buffer_size] = '\0';

	std::string ret = buffer;
	
	delete [] buffer;
	
	return ret;
}

OSStatus StringToEditTextConrol(WindowRef &m_window, const ControlID &kControl, const std::string &m_str)
{
	OSStatus result = noErr;
    ControlHandle tempControl;
 	
	if(NULL == m_window){
		throw __FILE__ ":StringToEditTextConrol: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":StringToEditTextConrol: Unable to get control";
    }
    
	result = SetControlData( tempControl, kControlEntireControl, kControlEditTextTextTag, 
		m_str.size(), m_str.c_str() );
	
	return result;
}

OSStatus IntToEditTextConrol(WindowRef &m_window, const ControlID &kControl, const int &m_int)
{
	OSStatus result = noErr;
    ControlHandle tempControl;
 	
	if(NULL == m_window){
		throw __FILE__ ":IntToEditTextConrol: NULL window reference";
	}
	
	result = GetControlByID(m_window, &kControl, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":IntToEditTextConrol: Unable to get control";
    }
    
	std::stringstream ss;
	
	ss << m_int;
	
	result = SetControlData( tempControl, kControlEntireControl, kControlEditTextTextTag, 
		ss.str().size(), ss.str().c_str() );
	
	return result;
}

bool HybridizePrimers(GenomePage *ptrGP)
{
    HIViewRef textView;
    TXNObject obj = NULL;
    TXNOffset start_offset, stop_offset;
	float strand;
	float salt;
	float tm;
	std::string primer1;
	std::string primer2;
	int exact_match;
	int min_amplicon_len;
	int max_amplicon_len;
	
    if(ptrGP == NULL){
        return false;
    }
    
    // Hybridization parameters
	strand = EditTextConrolToFloat(hybrid_window, kHybridStrandEdit);
	salt = EditTextConrolToFloat(hybrid_window, kHybridSaltEdit);
	tm = EditTextConrolToFloat(hybrid_window, kHybridTmEdit);
	exact_match = EditTextConrolToInt(hybrid_window, kHybridExactEdit);
	min_amplicon_len = EditTextConrolToInt(hybrid_window, kHybridMinAmpliconEdit);
	max_amplicon_len = EditTextConrolToInt(hybrid_window, kHybridMaxAmpliconEdit);
	primer1 = EditTextConrolToString(hybrid_window, kHybridPrimer1Edit);
	primer2 = EditTextConrolToString(hybrid_window, kHybridPrimer2Edit);
    
	if(exact_match < 0){
		throw "Please enter 0 or more bases for the exact 3' match";
	}
	
	if(min_amplicon_len < 0){
		throw "Please enter 0 or more bases for the minimum amplicon length";
	}
	
	if(max_amplicon_len < 0){
		throw "Please enter 0 or more bases for the maximum amplicon length";
	}
	
	if(max_amplicon_len < min_amplicon_len){
		throw "max amplicon length must be greater than the min length";
	}
	
    std::string search_results;
	
    try{
		SetThemeCursor(kThemeWatchCursor);
		
        search_results = ptrGP->search_primers(primer1, primer2,
			salt, strand, tm, exact_match, min_amplicon_len, max_amplicon_len);
    }
    catch(const char *error){
        search_results = error;
    }
	
	SetThemeCursor(kThemeArrowCursor);
	
    // Show the search results
    HIViewFindByID(HIViewGetRoot(hybrid_window), kHybridResultsEdit, &textView);
    
    obj = HITextViewGetTXNObject(textView);
    
    TXNSelectAll(obj);
    TXNGetSelection(obj, &start_offset, &stop_offset);
    
    // *Cludge* to get the window to show scroll bars for long lines that don't
    // fill the text window
    search_results = search_results + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";

    TXNSetData(obj, kTXNTextData, search_results.c_str(), search_results.size(), start_offset, stop_offset);
    
    // Display the upper left hand corner of text
    TXNSelectAll(obj);
    TXNShowSelection(obj, false);
    TXNSetSelection(obj, 0, 0);
    
    return true;
}

bool HybridizeProbe(GenomePage *ptrGP)
{
	HIViewRef textView;
    TXNObject obj = NULL;
    TXNOffset start_offset, stop_offset;
	float strand;
	float salt;
	float tm;
	std::string probe;
	
    if(ptrGP == NULL){
        return false;
    }
    
    // Hybridization parameters	
	strand = EditTextConrolToFloat(hybrid_window, kHybridStrandEdit);
	salt = EditTextConrolToFloat(hybrid_window, kHybridSaltEdit);
	tm = EditTextConrolToFloat(hybrid_window, kHybridTmEdit);
	probe = EditTextConrolToString(hybrid_window, kHybridProbeEdit);

    std::string search_results;
    
    try{
		SetThemeCursor(kThemeWatchCursor);
        search_results = ptrGP->search_probe(probe, salt, strand, tm);
    }
    catch(const char *error){
        search_results = error;
    }
	
	SetThemeCursor(kThemeArrowCursor);
	
    // Show the search results
    HIViewFindByID(HIViewGetRoot(hybrid_window), kHybridResultsEdit, &textView);
    
    obj = HITextViewGetTXNObject(textView);
    
    TXNSelectAll(obj);
    TXNGetSelection(obj, &start_offset, &stop_offset);
    
    // *Cludge* to get the window to show scroll bars for long lines that don't
    // fill the text window
    search_results = search_results + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";

    TXNSetData(obj, kTXNTextData, search_results.c_str(), search_results.size(), start_offset, stop_offset);
    
    // Display the upper left hand corner of text
    TXNSelectAll(obj);
    TXNShowSelection(obj, false);
    TXNSetSelection(obj, 0, 0);
    
    return true;
}

bool MOLigoPrimers(GenomePage *ptrGP)
{
    HIViewRef textView;
    TXNObject obj = NULL;
    TXNOffset start_offset, stop_offset;
	float strand;
	float salt;
	float tm;
	std::string primer1;
	std::string primer2;
	int exact_match;
	
    if(ptrGP == NULL){
        return false;
    }
    
    // Hybridization parameters
	strand = EditTextConrolToFloat(moligo_window, kMoligoStrandEdit);
	salt = EditTextConrolToFloat(moligo_window, kMoligoSaltEdit);
	tm = EditTextConrolToFloat(moligo_window, kMoligoTmEdit);
	exact_match = EditTextConrolToInt(moligo_window, kMoligoExactEdit);
	primer1 = EditTextConrolToString(moligo_window, kMoligoPrimer1Edit);
	primer2 = EditTextConrolToString(moligo_window, kMoligoPrimer2Edit);
    
	if(exact_match < 0){
		throw "Please enter 0 or more bases for the exact match about the ligation site";
	}
	
    std::string search_results;
    
    try{
		SetThemeCursor(kThemeWatchCursor);
		
        search_results = ptrGP->search_MOL_primers(primer1, primer2,
			salt, strand, tm, exact_match);
    }
    catch(const char *error){
        search_results = error;
    }
	
	SetThemeCursor(kThemeArrowCursor);
	
    // Show the search results
    HIViewFindByID(HIViewGetRoot(moligo_window), kMoligoResultsEdit, &textView);
    
    obj = HITextViewGetTXNObject(textView);
    
    TXNSelectAll(obj);
    TXNGetSelection(obj, &start_offset, &stop_offset);
    
    // *Cludge* to get the window to show scroll bars for long lines that don't
    // fill the text window
    search_results = search_results + "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";

    TXNSetData(obj, kTXNTextData, search_results.c_str(), search_results.size(), start_offset, stop_offset);
    
    // Display the upper left hand corner of text
    TXNSelectAll(obj);
    TXNShowSelection(obj, false);
    TXNSetSelection(obj, 0, 0);
    
    return true;
}

bool PickPrimers(GenomePage *ptrGP)
{
 	float strand;
	float salt;
	float min_primer_tm;
	float max_hairpin_tm;
	int primer_len;
	int run_len;
	float min_gc;
	float max_gc;
	
	unsigned int mask = 0;
	
    if(ptrGP == NULL){
        return false;
    }
    
    // Primer picking parameters	
	strand = EditTextConrolToFloat(primer_window, kPrimerStrandEdit);
	salt = EditTextConrolToFloat(primer_window, kPrimerSaltEdit);
	min_primer_tm = EditTextConrolToFloat(primer_window, kPrimerTmEdit);
	max_hairpin_tm = EditTextConrolToFloat(primer_window, kPrimerHairpinTmEdit);
	min_gc = EditTextConrolToFloat(primer_window, kPrimerMinGCEdit);
	max_gc = EditTextConrolToFloat(primer_window, kPrimerMaxGCEdit);

	primer_len = EditTextConrolToInt(primer_window, kPrimerLengthEdit);
	run_len = EditTextConrolToInt(primer_window, kPrimerRunLengthEdit);
    
	
	// Set the options mask
	if(CheckboxConrolToBool(primer_window, kPrimerPoly3GCCheck) == true){
		mask |= POLY_3_GC;
	}
	
	if(CheckboxConrolToBool(primer_window, kPrimerMulti5GCCheck) == true){
		mask |= MULTI_5_GC;
	}
	
	if(CheckboxConrolToBool(primer_window, kPrimerRunCheck) == true){
		mask |= NO_POLY_RUNS;
	}
	
	if(CheckboxConrolToBool(primer_window, kPrimerGCCheck) == true){
		mask |= GC_CONTENT;
	}
	
	if(CheckboxConrolToBool(primer_window, kPrimerTCheck) == true){
		mask |= NO_3_T;
	}
	
    try{
		SetThemeCursor(kThemeWatchCursor);
		
        ptrGP->compute_primers(mask,
			primer_len, 
			run_len,
			salt, 
			strand, 
			min_gc,
			max_gc,
			min_primer_tm,
			max_hairpin_tm);
    }
    catch(const char *error){
		SetThemeCursor(kThemeArrowCursor);
		
        ErrorDialog(error);
		
		return false;
    }
	
	SetThemeCursor(kThemeArrowCursor);

    return true;
}

void DownloadSearch(GenomePage *ptrGP)
{
    // The user has pressed the search button.
    
    OSStatus result = noErr;
    ControlHandle   tempControl;
    
    char *buffer = NULL;
    Size buffer_size = -1;
    
    if(ptrGP == NULL){
        return;
    }
        
    // First, extact the search string
    result = GetControlByID(download_window, &kDownloadSearchString, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to get search string control";
    }
    
    // Get the size of the buffer
    result = GetControlDataSize(tempControl, kControlEntireControl, kControlEditTextTextTag, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to get search string size";
    }

    buffer = new char[buffer_size + 1];
    
    if(buffer == NULL){
        throw __FILE__ ":DownloadSearch: Unable to allocate buffer!";
    }
    
    result = GetControlData(tempControl, kControlEntireControl, kControlEditTextTextTag, buffer_size, buffer, &buffer_size);
    
    if(result != noErr){
        throw __FILE__ ":FindSeq: Unable to get search string data";
    }
    
    // Add a string terminator
    buffer[buffer_size] = '\0';

    // Second, extract the Search key value
    result = GetControlByID(download_window, &kDownloadSearchKey, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to get search key control";
    }
    
    switch(GetControl32BitValue(tempControl)){
        case 1:
                docsum_list = docsum_list_from_search_key(buffer, 
                        ENTREZ_GENOME_ACCESSION);
                break;
        case 2:
                docsum_list = docsum_list_from_search_key(buffer, 
                        ENTREZ_GENOME_GI);
                break;
        case 3:
                docsum_list = docsum_list_from_search_key(buffer, 
                        ENTREZ_GENOME_TAXA_NAME);
                break;
        case 4:
                docsum_list = docsum_list_from_search_key(buffer, 
                        ENTREZ_GENOME_USER_DEFINED);
                break;			
        default:
                throw "Unknown download search key!";
    };

    if(buffer){
        delete [] buffer;
    }

    // Now, populate the results list
    result = GetControlByID(download_window, &kDownloadSearchResults, &tempControl);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to get search results control";
    }
    
    // Flush any existing entries in the data browser
    result = RemoveDataBrowserItems(tempControl, kDataBrowserNoItem, 0, NULL, kDataBrowserItemNoProperty);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to clear search results control";
    }
    
    std::list< std::pair<unsigned int, std::string> >::iterator iter;
    unsigned int list_count;
    
    for(list_count = 1, iter = docsum_list.begin();iter != docsum_list.end();iter++, list_count++){
        result = AddDataBrowserItems(tempControl, kDataBrowserNoItem, 1, 
			(const DataBrowserItemID*)(&list_count), kDataBrowserItemIsSelectableProperty);
        
        if(result != noErr){
            throw __FILE__ ":DownloadSearch: Unable to get add item to search results control";
        }
    }
    
}

bool DownloadLoad(GenomePage *ptrGP, const bool &m_save_file)
{
    // Load the users selection into the active window
    OSStatus        result = noErr;
    ControlHandle   browser;
    std::string file_name;
	
    if(ptrGP == NULL){
        return false;
    }
    
	// Do we need to prompt the user for a filename?
	if(m_save_file){
		if(GetSaveFileName(file_name) == false){
			return false;
		}
	}
	
    result = GetControlByID(download_window, &kDownloadSearchResults, &browser);
    
    if(result != noErr){
        throw __FILE__ ":DownloadSearch: Unable to get search results control";
    }
    
    std::list< std::pair<unsigned int, std::string> >::iterator iter;
    unsigned int list_count;
	unsigned int num_selected = 0;
    
	for(list_count = 1, iter = docsum_list.begin();iter != docsum_list.end();iter++){
        DataBrowserItemState itemState = 0L;
        
        result = GetDataBrowserItemState(browser, list_count, &itemState);
        
        if(result != noErr){
            throw __FILE__ ":DownloadLoad: Unable to get search results item";
        }
        
        if(itemState & kDataBrowserItemIsSelected){
			num_selected++;
		}
	}
	
	if( m_save_file && (num_selected != 1) ){
		ErrorDialog("Please select one and only one item to load and save");
		return false;
	}
	
    for(list_count = 1, iter = docsum_list.begin();iter != docsum_list.end();iter++, list_count++){
        DataBrowserItemState itemState = 0L;
        
        result = GetDataBrowserItemState(browser, list_count, &itemState);
        
        if(result != noErr){
            throw __FILE__ ":DownloadLoad: Unable to get search results item";
        }
        
        if(itemState & kDataBrowserItemIsSelected){
            // This item was selected
            try{
				bool ret = true;
				
                uout << "Downloading GI " << iter->first << " ..." << uout_endl;
				SetThemeCursor(kThemeWatchCursor);
				
				if(m_save_file){
					ret = ptrGP->load_and_save(iter->first, file_name);
				}
				else{
					ret = ptrGP->load(iter->first);
				}
				
				SetThemeCursor(kThemeArrowCursor);
				
				if(ret == false){
				
					ErrorDialog("Did not read any sequence data for this genome. "
						"Try downloading directly from the NCBI");
				}
            }
			catch(const char *error){
				SetThemeCursor(kThemeArrowCursor);
				
				ErrorDialog(error);
			}
            catch(...){
				SetThemeCursor(kThemeArrowCursor);
				
                ErrorDialog("An unhandled error has occured", "Please report this error to jgans@lanl.gov");
            }
        }
    }
    
    return true;
}

#pragma mark ---- Carbon Event Handling ----

pRecContext GetCurrentContextInfo (WindowRef window)
{
	if(NULL == window){  // HID use this path
		window = FrontNonFloatingWindow();
	}
	
	if(window){
		return (pRecContext) GetWRefCon (window);
	}

	return NULL;
}

// ---------------------------------

static pascal OSErr	ReceiveDroppedFiles(
	const AppleEvent 	*inputEvent,
	AppleEvent 	*outputEvent,
	SInt32		handlerRefCon)
{
	OSStatus err = noErr;
	WindowRef window = NULL;
	pRecContext pContextInfo = NULL;
	AEDescList docList;
	long itemsInList;
	
	// Is there a window that can receive this data?
	// Append to the current window if available, create a new window
	// if none exists.
	window = FrontNonFloatingWindow();
	
	if(window == NULL){
		// Create a new window
		window = createNewWindow();
		
		if(window == NULL){
			// Couldn't create a new window!
			return err; // Just return and don't report an error!
		}
	}
	
	pContextInfo = GetCurrentContextInfo (window);
	
	if(pContextInfo == NULL){
		return err; // Just return and don't report an error!
	}
	
	// Extract the filenames from the inputEvent
	err = AEGetParamDesc( inputEvent, keyDirectObject, typeAEList, &docList);
	
	if(err != noErr){
		return err;
	}
	
	// The number of files
	err = AECountItems( &docList, &itemsInList);
	
	if(err != noErr){
		return err;
	}
	
	bool load_success = true;
	
	// Iterate through each file
	for(long index = 1; index <= itemsInList; index++){
		try{
			AEKeyword key_word;
			AEDesc desc_result;
			
			// Request the output in FSRef format instead of FSSpec format.
			// This will allow the use of the FSRefMakePath function to extract the
			// filename as a plain old vanilla C-style string.
			err = AEGetNthDesc (
				&docList,
				index,
				typeFSRef,
				&key_word,
				&desc_result);
				
			if(err != noErr){
				throw __FILE__ "Error in AEGetNthDesc";
			}
			
			// Get the file name.
			FSRef theRef;
			
			err = AEGetDescData(&desc_result, &theRef, sizeof(FSRef));
			
			if(err != noErr){
				throw __FILE__ "Error in AEGetDescData";
			}
			
			// Convert the FSRef to a path name
			char filename[512];
			
			FSRefMakePath(&theRef, (UInt8*)filename, 512);
			
			SetThemeCursor(kThemeWatchCursor);
			
			// Open the file. Whew!
			if(pContextInfo->ptrGP->load(filename, file_type(filename)) == false){
				
				SetThemeCursor(kThemeArrowCursor);
				
				ErrorDialog("Did not read any sequence data for this genome. "
						"Try downloading directly from the NCBI");
			}
			
			err = AEDisposeDesc(&desc_result);
		}
		catch(...){
			// Set the error flag
			load_success = false;
		}
		
		SetThemeCursor(kThemeArrowCursor);
	}
	
	if(load_success == false){
		// Let the user know that there were errors loading the requested data
		ErrorDialog("Errors were encountered while loading data");
	}
	
	// Update the display
	Rect rectPort = {0,0,0,0};
                         
	GetWindowPortBounds (window, &rectPort);
	InvalWindowRect (window, &rectPort);
	
	return err;
}

// Open a dialog box to recive user filenames to open
bool OpenFiles(WindowRef window, pRecContext pContextInfo)
{
    NavDialogCreationOptions inOptions;
    NavTypeListHandle inTypeList;
    NavDialogRef outDialog;
    NavReplyRecord outReply;
    OSStatus result = noErr;
    NavUserAction action;
    
    // Do we have a valid data structure to read into?
    if(!pContextInfo || !pContextInfo->ptrGP){
        return false;
    }
    
    // Set the dialog options
    result = NavGetDefaultDialogCreationOptions(&inOptions);

    if(noErr != result){
        throw __FILE__ ":OpenFiles: Error in NavGetDefaultDialogCreationOptions";
    }
    
    inTypeList = (NavTypeListHandle)GetResource('open', 128);
    
    result = NavCreateGetFileDialog(&inOptions, 
        inTypeList, 
        NULL, 
        NULL,
        NULL,
        NULL,
        &outDialog);

    if(noErr != result){
        throw __FILE__ ":OpenFiles: Error in NavCreateGetFileDialog";
    }
    
    result = NavDialogRun(outDialog);
    
    if(noErr != result){
        throw __FILE__ ":OpenFiles: Error in NavDialogRun";
    }
    
    action = NavDialogGetUserAction(outDialog);

    // Do we need to process file name(s)?    
    if((action != kNavUserActionCancel) && (action != kNavUserActionNone)){
    
        result = NavDialogGetReply(outDialog, &outReply);
    
        if(noErr != result){
            throw __FILE__ ":OpenFiles: Error in NavDialogGetReply";
        }
        
        // Is this data structure valid?
        if(outReply.validRecord){
            SInt32 file_count;
            
            result = AECountItems(&outReply.selection, &file_count);
            
            if(result != noErr){
                throw __FILE__ ":OpenFiles: Error in AECountItems";
            }
            
			try{
				for(int i = 1;i <= file_count;i++){
					AEKeyword key_word;
					AEDesc desc_result;
					
					// Request the output in FSRef format instead of FSSpec format.
					// This will allow the use of the FSRefMakePath function to extract the
					// filename as a plain old vanilla C-style string.
					result = AEGetNthDesc (
						&outReply.selection,
						i,
						typeFSRef,
						&key_word,
						&desc_result);
						
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEGetNthDesc";
					}
					
					// Get the file name.
					FSRef theRef;
					
					result = AEGetDescData(&desc_result, &theRef, sizeof(FSRef));
					
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEGetDescData";
					}
					
					// Convert the FSRef to a path name
					char filename[512];
					
					FSRefMakePath(&theRef, (UInt8*)filename, 512);
					
					SetThemeCursor(kThemeWatchCursor);
					
					// Open the file. Whew!
					if(pContextInfo->ptrGP->load(filename, file_type(filename)) == false){
					
						SetThemeCursor(kThemeArrowCursor);
						
						ErrorDialog("Did not read any sequence data for this genome. "
							"Try downloading directly from the NCBI");
					}
					
					result = AEDisposeDesc(&desc_result);
					
					if(result != noErr){
						throw __FILE__ ":OpenFiles: Error in AEDisposeDesc";
					}
					
					SetThemeCursor(kThemeArrowCursor);
				}
            }
			catch(const char *error){
				SetThemeCursor(kThemeArrowCursor);
				
				ErrorDialog(error);
			}
			catch(...){
				SetThemeCursor(kThemeArrowCursor);
				
				ErrorDialog("Unhandled error opening file", 
					"Please report this error to jgans@lanl.gov");
			}
			
            result = NavDisposeReply(&outReply);
    
            if(noErr != result){
                throw __FILE__ ":OpenFiles: Error in NavDisposeReply";
            }

        }
    }
    
    NavDialogDispose(outDialog);
    
    if (inTypeList != NULL){
        ReleaseResource( (char**)inTypeList );
    }
    
    return true;
}

// Open a dialog box to recive user genome plot filenames to open
bool OpenFilesGenomePlot(WindowRef window, pRecContext pContextInfo)
{
    NavDialogCreationOptions inOptions;
    NavTypeListHandle inTypeList;
    NavDialogRef outDialog;
    NavReplyRecord outReply;
    OSStatus result = noErr;
    NavUserAction action;
    
    // Do we have a valid data structure to read into?
    if(!pContextInfo || !pContextInfo->ptrGP){
        return false;
    }
    
    // Set the dialog options
    result = NavGetDefaultDialogCreationOptions(&inOptions);

    if(noErr != result){
        throw __FILE__ ":OpenFilesGenomePlot: Error in NavGetDefaultDialogCreationOptions";
    }
    
    inTypeList = (NavTypeListHandle)GetResource('open', 128);
    
    result = NavCreateGetFileDialog(&inOptions, 
        inTypeList, 
        NULL, 
        NULL,
        NULL,
        NULL,
        &outDialog);

    if(noErr != result){
        throw __FILE__ ":OpenFilesGenomePlot: Error in NavCreateGetFileDialog";
    }
    
    result = NavDialogRun(outDialog);
    
    if(noErr != result){
        throw __FILE__ ":OpenFilesGenomePlot: Error in NavDialogRun";
    }
    
    action = NavDialogGetUserAction(outDialog);

    // Do we need to process file name(s)?    
    if((action != kNavUserActionCancel) && (action != kNavUserActionNone)){
    
        result = NavDialogGetReply(outDialog, &outReply);
    
        if(noErr != result){
            throw __FILE__ ":OpenFilesGenomePlot: Error in NavDialogGetReply";
        }
        
        // Is this data structure valid?
        if(outReply.validRecord){
            SInt32 file_count;
            
            result = AECountItems(&outReply.selection, &file_count);
            
            if(result != noErr){
                throw __FILE__ ":OpenFilesGenomePlot: Error in AECountItems";
            }
            
            for(int i = 1;i <= file_count;i++){
                AEKeyword key_word;
                AEDesc desc_result;
                
                // Request the output in FSRef format instead of FSSpec format.
                // This will allow the use of the FSRefMakePath function to extract the
                // filename as a plain old vanilla C-style string.
                result = AEGetNthDesc (
                    &outReply.selection,
                    i,
                    typeFSRef,
                    &key_word,
                    &desc_result);
                    
                if(result != noErr){
                    throw __FILE__ ":OpenFilesGenomePlot: Error in AEGetNthDesc";
                }
                
                // Get the file name.
                FSRef theRef;
                
                result = AEGetDescData(&desc_result, &theRef, sizeof(FSRef));
                
                if(result != noErr){
                    throw __FILE__ ":OpenFilesGenomePlot: Error in AEGetDescData";
                }
                
                // Convert the FSRef to a path name
                char filename[512];
                
                FSRefMakePath(&theRef, (UInt8*)filename, 512);
                
                // Open the file. Whew!
				SetThemeCursor(kThemeWatchCursor);
				
				try{
					pContextInfo->ptrGP->read_genome_plot(filename);
					SetThemeCursor(kThemeArrowCursor);
                }
				catch(const char *error){
					SetThemeCursor(kThemeArrowCursor);
					ErrorDialog(error);
				}
				catch(...){
					SetThemeCursor(kThemeArrowCursor);
					ErrorDialog("Unable to read genome plot");
				}
				
                result = AEDisposeDesc(&desc_result);
                
                if(result != noErr){
                    throw __FILE__ ":OpenFilesGenomePlot: Error in AEDisposeDesc";
                }
            }
            
            result = NavDisposeReply(&outReply);
    
            if(noErr != result){
                throw __FILE__ ":OpenFilesGenomePlot: Error in NavDisposeReply";
            }

        }
    }
    
    NavDialogDispose(outDialog);
    
    if (inTypeList != NULL){
        ReleaseResource( (char**)inTypeList );
    }
    
    return true;
}


// Open a dialog box to prompt the user for a filename.
bool ImportAnnotation(WindowRef window, pRecContext pContextInfo)
{
	if( (pContextInfo->ptrGP == NULL) || (pContextInfo->ptrGP->num_display() == 0) ){
		reportError("Please open a genome before importing additional annotations");
		return false;
	}
	
	std::list<std::string> filenames;
	
	if( pContextInfo->ptrGP && GetMultiFiles(filenames) ){
	
		std::list<std::string>::iterator iter;
		
		for(iter = filenames.begin();iter != filenames.end();iter++){
			pContextInfo->ptrGP->import_annotation(*iter);
		}
		
		return true;
	}
            
    return false;
}

// Open a dialog box to prompt the user for a filename.
// Export the currently selected genomes to this file.
bool ExportGenome(WindowRef window, pRecContext pContextInfo)
{
	if( !pContextInfo || !(pContextInfo->ptrGP) || (pContextInfo->ptrGP->num_display() == 0) ){
	
		reportError("Please open a genome before attempting to export a genome");
		return false;
	}
	
	ControlRef control_ref;
	
	// First, make sure the user has selected at least one type 
	// of annotation to export
	pContextInfo->ptrGP->clear_annotations_to_export();
	
	GetControlByID(export_window, &kExportAnnotGene, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::GENE);
	}
	
	GetControlByID(export_window, &kExportAnnotPseudo, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::PSEUDO_GENE);
	}
	
	GetControlByID(export_window, &kExportAnnotCDS, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::CDS);
	}
	
	GetControlByID(export_window, &kExportAnnotRNA, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::RNA);
	}
	
	GetControlByID(export_window, &kExportAnnottRNA, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::tRNA);
	}
	
	GetControlByID(export_window, &kExportAnnotMisc, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::IMP);
	}
	
	GetControlByID(export_window, &kExportAnnotprimer, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::PRIMER);
	}
	
	GetControlByID(export_window, &kExportAnnotUser, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::USER);
	}
	
	GetControlByID(export_window, &kExportAnnotIntergenic, &control_ref);
	if(GetControlValue(control_ref) == kControlCheckBoxCheckedValue){
		pContextInfo->ptrGP->add_annotation_to_export(GeneAnnotation::NONE);
	}
	
	if(pContextInfo->ptrGP->num_annotations_to_export() == 0){
		ErrorDialog("Please select at least on annotation type to export");
		return false;
	}
	
	GetControlByID(export_window, &kExportFormat, &control_ref);
	
	switch( GetControlValue(control_ref) ){
		case 1: // Fasta (entire genome)
			if(pContextInfo->ptrGP->set_export_format(GenomePage::EXPORT_FASTA_GENOME) == false){
			
				ErrorDialog("Please specify a valid export format");
				return false;
			}
			break;
		case 2: // Fasta (DNA)
			if(pContextInfo->ptrGP->set_export_format(GenomePage::EXPORT_FASTA_DNA) == false){
			
				ErrorDialog("Please specify a valid export format");
				return false;
			}
			break;
		case 3: // Fasta (amino acid)
			if(pContextInfo->ptrGP->set_export_format(GenomePage::EXPORT_FASTA_PROTEIN) == false){
			
				ErrorDialog("Please specify a valid export format");
				return false;
			}
			break;
		case 4: // GBK (genbank annotation)
			if(pContextInfo->ptrGP->set_export_format(GenomePage::EXPORT_GBK) == false){
			
				ErrorDialog("Please specify a valid export format");
				return false;
			}
			break;
		case 5: // PTT (protein table)
			if(pContextInfo->ptrGP->set_export_format(GenomePage::EXPORT_PTT) == false){
			
				ErrorDialog("The PTT format can only store a single genome");
				return false;
			}
			break;
		default:
			throw "Unknown export format!";
	};
	
	std::string file_name;
	
	if( pContextInfo->ptrGP && GetSaveFileName(file_name) ){
		
		try{
			SetThemeCursor(kThemeWatchCursor);
			
			pContextInfo->ptrGP->export_genomes(file_name);
		}
		catch(const char *error){
		
			SetThemeCursor(kThemeArrowCursor);
			ErrorDialog(error);
			return false;
		}
		catch(...){
			SetThemeCursor(kThemeArrowCursor);
			ErrorDialog("An export error has occured");
			return false;
		}
	}
	
	SetThemeCursor(kThemeArrowCursor);
    return true;
}


// Open a dialog box to prompt the user for a filename.
// Export the current view in postscript to this file
bool ExportPS(WindowRef window, pRecContext pContextInfo)
{
	std::string file_name;
	
	if( pContextInfo->ptrGP && GetSaveFileName(file_name) ){
		pContextInfo->ptrGP->render_ps(file_name);
		
		return true;
	}
            
    return false;
}

// Open a dialog box to prompt the user for a filename.
// Export the current view in GIF to this file
bool ExportGIF(WindowRef window, pRecContext pContextInfo)
{
	std::string file_name;
	
	if( pContextInfo->ptrGP && GetSaveFileName(file_name) ){
		pContextInfo->ptrGP->render_gif(file_name);
		
		return true;
	}
            
    return false;
}

// Update the selected and allowed annotations on the export window
// in response to a format change
void UpdateExportWindow()
{
	ControlRef control_ref;
	
	// Which format has been selected?
	GetControlByID(export_window, &kExportFormat, &control_ref);
	
	switch( GetControlValue(control_ref) ){
		case 1: // Fasta (entire genome)
			// Turn on and deactivate all annotations
			GetControlByID(export_window, &kExportAnnotGene, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotPseudo, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotCDS, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotRNA, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnottRNA, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotMisc, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotprimer, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotUser, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotIntergenic, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxCheckedValue);
			DeactivateControl(control_ref);
			
			break;
		case 2: // Fasta (DNA)
		case 4: // GBK (genbank annotation)
		case 5: // PTT (protein table)
			// Activate all annotations and turn off intergenic space
			GetControlByID(export_window, &kExportAnnotGene, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotPseudo, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotCDS, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotRNA, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnottRNA, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotMisc, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotprimer, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotUser, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotIntergenic, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			break;
		case 3: // Fasta (amino acid)
			// Activate all annotations and turn off non-amino acid annotations
			GetControlByID(export_window, &kExportAnnotGene, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotPseudo, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotCDS, &control_ref);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotRNA, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnottRNA, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotMisc, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotprimer, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotUser, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			
			GetControlByID(export_window, &kExportAnnotIntergenic, &control_ref);
			SetControlValue(control_ref, kControlCheckBoxUncheckedValue);
			ActivateControl(control_ref);
			break;
		default:
			throw "Unknown export format!";
	};
}

// Handles global non-system mouse events for GL windows as follows:
// primary button
// secondary button
// tertiary button
// wheel	
static OSStatus handleWindowMouseEvents (EventHandlerCallRef myHandler, EventRef event)
{
    WindowRef           window = NULL;
    pRecContext         pContextInfo = NULL;
    OSStatus            result = eventNotHandledErr;
    UInt32              kind = GetEventKind (event);
    EventMouseButton	button = 0;
    HIPoint             location = {0.0f, 0.0f};
    UInt32              modifiers = 0;
    UInt32              ClkCount = 0;	
    long                wheelDelta = 0; 
    bool                update_display = false;
    Rect                rectPort = {0,0,0,0};
    Rect                rectWindow = {0,0,0,0};
    float               title_bar_size = 0.0f;
    CFStringRef         title = NULL;
    std::stringstream   sout;
    
    // Mac OS X v10.1 and later
    // should this be front window???
    GetEventParameter(event, kEventParamWindowRef, typeWindowRef, NULL, sizeof(WindowRef), NULL, &window);
	
    if (window){
        pContextInfo = GetCurrentContextInfo (window);
        
        // Get BOTH the port and window bounds for invalidation and mouse location
        // offset (we need to subtract the size of the title bar).
        GetWindowPortBounds (window, &rectPort);
        GetWindowBounds (window, kWindowStructureRgn, &rectWindow);
        
        title_bar_size = (rectWindow.bottom - rectWindow.top) - (rectPort.bottom - rectPort.top);
    }
        
    if (!pContextInfo){
        return result; // not an application GLWindow so do not process (there is an exception)
    }
	
	// Set the cursor
	Cursor cursor_arrow;
	
	SetCursor(GetQDGlobalsArrow(&cursor_arrow));
	
    result = CallNextEventHandler(myHandler, event);
	
	// Protect this event loop in a try-catch
	try{
	
    if (eventNotHandledErr == result) { 
        // only handle events not already handled (prevents wierd resize interaction)
        switch (kind) {
            // start trackball, pan, or dolly
            case kEventMouseDown:
                GetEventParameter(event, kEventParamMouseButton, typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &button);
                GetEventParameter(event, kEventParamWindowMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
                GetEventParameter(event, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(UInt32), NULL, &modifiers);
                GetEventParameter(event, kEventParamClickCount, typeUInt32, NULL, sizeof(UInt32), NULL, &ClkCount);
                
                location.y -= title_bar_size;
                
                if ((button == kEventMouseButtonPrimary) && (modifiers & cmdKey)) {
					// Be Apple specific and use the command key (cmdKey) not the control key (conrolKey)
					update_display = pContextInfo->ptrGP->l_button_ctrl_down((int)location.x, (int)location.y);
                } 
                else if ((button == kEventMouseButtonPrimary) && (modifiers & shiftKey)) {
				
					// Shift-clicking saves the current base in the translation buffer and, when the buffer is full,
					// displayed the selected text.
					std::string info;
					
					update_display = pContextInfo->ptrGP->l_button_shift_down((int)location.x, (int)location.y, info);
					
					if(info.empty() == false){
						 DisplayInfo(info);
					}
				} 
                else if (button == kEventMouseButtonPrimary)  { // Pan left/right
                    if(ClkCount == 1){
                        update_display = pContextInfo->ptrGP->l_button_down((int)location.x, (int)location.y);
                    }
                    else{
						// Double clicking displays gene/genome information in a dialog
                        std::string info;
                        
                        update_display = pContextInfo->ptrGP->l_button_dbl_down((int)location.x, (int)location.y, info);
                        
                        if(info.empty() == true){
                            info = "No information to display";
                        }

                        DisplayInfo(info);
                    }    
                }
				else if(button == kEventMouseButtonSecondary){ // Zoom
					update_display = pContextInfo->ptrGP->r_button_down((int)location.x, (int)location.y);
				}
				else if(button == kEventMouseButtonTertiary){ // Scroll up/down
					update_display = pContextInfo->ptrGP->m_button_down((int)location.x, (int)location.y);
				}
				
                break;
            case kEventMouseUp:
                GetEventParameter(event, kEventParamMouseButton, typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &button);
				GetEventParameter(event, kEventParamWindowMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
				
				if(button == kEventMouseButtonPrimary){ // Pan left/right
					update_display = pContextInfo->ptrGP->l_button_up((int)location.x, (int)location.y);
				}
				else if(button == kEventMouseButtonSecondary){ // Zoom
					update_display = pContextInfo->ptrGP->r_button_up((int)location.x, (int)location.y);
				}
				else if(button == kEventMouseButtonTertiary){ // Scroll up/down
					update_display = pContextInfo->ptrGP->m_button_up((int)location.x, (int)location.y);
				}
				
                break;
			case kEventMouseWheelMoved:
				GetEventParameter(event, kEventParamWindowMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
                GetEventParameter(event, kEventParamMouseWheelDelta, typeLongInteger, NULL, sizeof(long), NULL, &wheelDelta);
                
				if (wheelDelta){
                        update_display = pContextInfo->ptrGP->mouse_wheel((int)location.x, (int)location.y, wheelDelta);
                }
				 
				break;
            case kEventMouseMoved:
            case kEventMouseDragged:
                GetEventParameter(event, kEventParamWindowMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
                
                location.y -= title_bar_size;
                
                update_display = pContextInfo->ptrGP->mouse_move((int)location.x, (int)location.y);
                
                // Display the current base in the title bar (always)
                sout << "base(" << pContextInfo->ptrGP->cur_base_loc() << ")";
				
				sout.precision(3);
				
				// Display the Tm (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_tm()){
					sout << ":Tm(" << pContextInfo->ptrGP->cur_tm() << ")";
				}
				
				// Display the %GC (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_gc()){
					sout << ":%GC(" << pContextInfo->ptrGP->cur_gc() << ")";
				}
				
				// Display the %AT (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_at()){
					sout << ":%AT(" << pContextInfo->ptrGP->cur_at() << ")";
				}
				
				// Display the GC skew (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_gc_skew()){
					sout << ":GC skew(" << pContextInfo->ptrGP->cur_gc_skew() << ")";
				}
				
				// Display the AT skew (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_at_skew()){
					sout << ":AT skew(" << pContextInfo->ptrGP->cur_at_skew() << ")";
				}
				
				// Display the Keto skew (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_keto_skew()){
					sout << ":keto(" << pContextInfo->ptrGP->cur_keto_skew() << ")";
				}
				
				// Display the Purine skew (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_purine_skew()){
					sout << ":purine(" << pContextInfo->ptrGP->cur_purine_skew() << ")";
				}
				
				// Display the genome plot (if requested by the user)
				if(pContextInfo->ptrGP->cur_show_genome_plot()){
					sout << ":user(" << pContextInfo->ptrGP->cur_genome_plot() << ")";
				}
				
				title = CFStringCreateWithCString(NULL, sout.str().c_str(), CFStringGetSystemEncoding());
                
                SetWindowTitleWithCFString(window, title);
                
                CFRelease(title);
                
                break;
            };
            
            result = noErr;
	}
        
	if(update_display){
		InvalWindowRect (window, &rectPort);
	}

	}
	catch(const char *error){
		ErrorDialog(error);
	}
	catch(...){
		ErrorDialog("An unhandled error has occured", "Please report this error to jgans@lanl.gov");
	}
	
	return result;
}

// ---------------------------------

// key input handler
static OSStatus handleKeyInput (EventHandlerCallRef myHandler, EventRef event, Boolean keyDown, void* userData)
{
        enum { kPageUp     = 0x74,
               kPageDown   = 0x79, 
               kArrowUp    = 0x7E,
               kArrowDown  = 0x7D,
               kArrowLeft  = 0x7B,
               kArrowRight = 0x7C,
               kEsc        = 0x35,
               kDel1       = 0x33,
               kDel2       = 0x75,
               kg          = 0x05,
               k1          = 0x12,
               k2          = 0x13,
               k3          = 0x14,
               k4          = 0x15,
               k5          = 0x17,
               k6          = 0x16,
               k7          = 0x1A,
               k8          = 0x1C,
               k9          = 0x19,
               k0          = 0x1D};
        
	WindowRef window = (WindowRef) userData;
	OSStatus result = eventNotHandledErr;
	UInt32 modifiers = 0;
	bool update_view = false;
        
	result = CallNextEventHandler(myHandler, event);
	
	// Protrect this event loop in a try-catch
	try{
	
	if (eventNotHandledErr == result) { 
            pRecContext pContextInfo = GetCurrentContextInfo (window);
            
            if(pContextInfo && pContextInfo->ptrGP) {
                UInt32 keyCode;
			
                GetEventParameter (event, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keyCode);
				GetEventParameter(event, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(UInt32), NULL, &modifiers);
				
                //fprintf(stderr, "KeyCode: %lu (0x%lX)\n", keyCode, keyCode);
                
                // handle keyboard input here
                if (keyDown) {
                    switch (keyCode) {
                        case kPageUp:
                            update_view = pContextInfo->ptrGP->key_page_up();
                            break;
                        case kPageDown:
                            update_view = pContextInfo->ptrGP->key_page_down();
                            break;
                        case kArrowUp:
							if(modifiers & shiftKey){
								update_view = pContextInfo->ptrGP->shift_selection_up();
							}
							else{
								update_view = pContextInfo->ptrGP->key_up_arrow();
							}
                            break;
                        case kArrowDown:
							if(modifiers & shiftKey){
								update_view = pContextInfo->ptrGP->shift_selection_down();
							}
							else{
								update_view = pContextInfo->ptrGP->key_down_arrow();
							}
                            break;
                        case kArrowLeft:
                            update_view = pContextInfo->ptrGP->key_left_arrow();
                            break;
                        case kArrowRight:
                            update_view = pContextInfo->ptrGP->key_right_arrow();
                            break;
                        case kEsc:
                            update_view = pContextInfo->ptrGP->clear_selection();
                            pContextInfo->ptrGP->jump_buffer_clear();
                            break;
						// Disable the keyboard short cut for deleting genomes. It's too easy
						// to accidentaly delete a genome by pressing the wrong key (wait until
						// undo functionality is added).
                        //case kDel1:
                        //case kDel2:
                        //    update_view = pContextInfo->ptrGP->delete_selection();
                        //    break;
                        case kg:
							if(modifiers & shiftKey){
								pContextInfo->ptrGP->jump_to();
								pContextInfo->ptrGP->jump_buffer_clear();
								update_view = true;
							}
                            break;
                        case k0:
                            pContextInfo->ptrGP->jump_buffer_push(0);
                            break;
                        case k1:
                            pContextInfo->ptrGP->jump_buffer_push(1);
                            break;
                        case k2:
                            pContextInfo->ptrGP->jump_buffer_push(2);
                            break;
                        case k3:
                            pContextInfo->ptrGP->jump_buffer_push(3);
                            break;
                        case k4:
                            pContextInfo->ptrGP->jump_buffer_push(4);
                            break;
                        case k5:
                            pContextInfo->ptrGP->jump_buffer_push(5);
                            break;
                        case k6:
                            pContextInfo->ptrGP->jump_buffer_push(6);
                            break;
                        case k7:
                            pContextInfo->ptrGP->jump_buffer_push(7);
                            break;
                        case k8:
                            pContextInfo->ptrGP->jump_buffer_push(8);
                            break;
                        case k9:
                            pContextInfo->ptrGP->jump_buffer_push(9);
                            break;                                    
                    }
                                
                    if(update_view){
                        Rect rectPort = {0,0,0,0};
                         
                        GetWindowPortBounds (window, &rectPort);
                        InvalWindowRect (window, &rectPort);
                        //drawGL(window, true);
                    }
                    
                    result = noErr;
                }
            }
	}
	
	}
	catch(const char *error){
		ErrorDialog(error);
	}
	catch(...){
		ErrorDialog("An unhandled error has occured", "Please report this error to jgans@lanl.gov");
	}
	
	return result;
}
		
// ---------------------------------

// window event handler
static pascal OSStatus windowEvtHndlr (EventHandlerCallRef myHandler, EventRef event, void* userData)
{
#pragma unused (userData)
    WindowRef			window = NULL;
    pRecContext			pContextInfo = NULL;
    Rect				rectPort = {0,0,0,0};
    int					width;
    int					height;
    OSStatus			result = eventNotHandledErr;
    UInt32				e_class = GetEventClass (event);
    UInt32				kind = GetEventKind (event);

	// Proetct this event function in a try-catch
	try{
	
	switch (e_class) {
		case kEventClassKeyboard:
			switch (kind) {
				case kEventRawKeyDown:
                                case kEventRawKeyRepeat:
					result = handleKeyInput  (myHandler, event, true, userData);
					break;
                                // Ignore key up messages for now
				//case kEventRawKeyUp:
				//	result = handleKeyInput  (myHandler, event, false, userData);
				//	break;
			}
			break;
		case kEventClassWindow:
			GetEventParameter(event, kEventParamDirectObject, typeWindowRef, NULL, sizeof(WindowRef), NULL, &window);
			switch (kind) {
				case kEventWindowCollapsing:
					pContextInfo = GetCurrentContextInfo (window);
					GetWindowPortBounds (window, &rectPort);
					drawGL (window, false); // in this case just draw content
					CompositeGLBufferIntoWindow( pContextInfo->aglContext, &rectPort, window);
					result = UpdateCollapsedWindowDockTile (window);
					break;
				case kEventWindowActivated: // called on click activation and initially
					// If an openGL window has grabbed the AGL context, we will need
					// to grab it back before making any openGL calls.
					pContextInfo = GetCurrentContextInfo (window);
					
					if(pContextInfo && pContextInfo->aglContext  && pContextInfo->ptrGP){
						// re-attach drawable to ensure context is updated
						aglSetCurrentContext (pContextInfo->aglContext);// Set the arrow cursor
						Cursor arrow_cursor;
						SetCursor(GetQDGlobalsArrow(&arrow_cursor));
					}
					
					// Update the preferences dialog if needed
					if(option_window && pContextInfo){
						UpdateOptionsDialog(pContextInfo);
					}
					
					break;
				// Note that the kEventWindowUpdate and kEventWindowDrawContent message are only sent
				// if the window has been configured to receive updates! This configuration is
				// done in the Nib editor -- look at the info for the main window.
				//case kEventWindowUpdate:
				case kEventWindowDrawContent: // will receive one prior to being shown...
					drawGL (window, true); // in this case just draw content
					break;
				case kEventWindowClose: // called when window is being closed (close box)
					HideWindow (window);
					pContextInfo = GetCurrentContextInfo (window);
					disposeGL (pContextInfo);
					DisposePtr ((Ptr) pContextInfo);
                                        SetWRefCon (window, NULL);
					break;
				case kEventWindowShown: // called on initial show (not on un-minimize)
					// draw content is received prior to this so if you handle that then not drawing required here
					if (window == FrontWindow ()){
						SetUserFocusWindow (window);
					}

					drawGL (window, true); // in this case just draw content
					break;
				case kEventWindowBoundsChanged: // called for resize and moves (drag)
					pContextInfo = GetCurrentContextInfo (window);	
					if (pContextInfo) { // if we have a vlaid context
						GetWindowPortBounds (window, &rectPort); // get the bounds and set the view rect
                                                
						width = rectPort.right - rectPort.left;
						height = rectPort.bottom - rectPort.top;
						
						resizeGL (pContextInfo, rectPort); // forces updateContext & viewport update if required
						
						// Are we increasing the window
						if ((height > pContextInfo->height) || (width > pContextInfo->width)) {
							
							// resize to update context virtual screen
							//resizeGL (pContextInfo, width, height); // forces updateContext & viewport update if required
							//resizeGL (pContextInfo, rectPort); // forces updateContext & viewport update if required
							
							if (pContextInfo->currentVS != aglGetVirtualScreen (pContextInfo->aglContext)) { // if virtual screen (i.e. renderer) changes
								pContextInfo->currentVS = aglGetVirtualScreen (pContextInfo->aglContext); // sync renderer
							}
						}
						//} else {
							// resize prior to update
							//resizeGL (pContextInfo, width, height); // forces updateContext & viewport update if required
							//resizeGL (pContextInfo, width, height); // forces updateContext & viewport update if required
						//}
                                                
						drawGL (window, true);// force immediate update to get live resize on shrink
					}
					break;
				case kEventWindowZoomed: // called when user clicks on zoom button (occurs after the window has been zoomed)
					// use this if you need to some special here as you always get a kEventWindowBoundsChanged event
					break;
			}
			break;
	}
	
	}
	catch(const char *error){
		ErrorDialog(error);
	}
	catch(...){
		ErrorDialog("An unhandled error has occured", "Please report this error to jgans@lanl.gov");
	}
	
    return result;
}

// ---------------------------------

// application level event handler
static pascal OSStatus appEvtHndlr (EventHandlerCallRef myHandler, EventRef event, void* userData)
{
#pragma unused (myHandler, userData)
    OSStatus result = eventNotHandledErr;
    Rect rectPort;
    pRecContext pContextInfo = NULL;
    WindowRef window = FrontNonFloatingWindow();
    UInt32 e_class = GetEventClass (event);
    UInt32 kind = GetEventKind (event);
    HICommand command;
    const float color_8_to_16 = 65535.0f/255.0f;
    const float color_16_to_8 = 1.0f/color_8_to_16;
    
    if (window) { // do we have a window?
        GetWindowPortBounds (window, &rectPort); // get bounds for later inval
        pContextInfo = GetCurrentContextInfo (window);
    }
    
	// Protect the entire event loop with a try-catch
	try{
	
	switch (e_class) {
		case kEventClassMouse:
			handleWindowMouseEvents (myHandler, event);
			break;
		case kEventClassCommand:
			switch (kind) {
				case kEventProcessCommand:
					
					GetEventParameter (event, kEventParamDirectObject, kEventParamHICommand, NULL, sizeof(command), NULL, &command); // get command
					switch (command.commandID) {
						case kHICommandNew:
							createNewWindow ();
							result = noErr;
							break;
						case kHICommandClose:
							if (window) {
								HideWindow (window);
								disposeGL (GetCurrentContextInfo(window));  // dump our structures
								DisposeWindow (window); // if it exists dump it
							}
							break;
						case kHICommandQuit:
							//memory will get dumped on exit, other callbacks/timers are for this process only
							break;
						case 'hscl':
							if(window){
								// Return the hoizontal scroll bar to its default position (i.e. "snap back")
								ControlRef control;
		
								GetControlByID(window, &kHorizontalScrollBar, &control);
								SetControl32BitValue(control, DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);

								if(pContextInfo && pContextInfo->ptrGP){
									pContextInfo->ptrGP->scroll_bar(DEFAULT_HORIZONTAL_SCROLL_BAR_LOC);
								}
							}
							
							break;
						case 'vscl':
							if(window){
								// Return the vertical scroll bar to its default position (i.e. "snap back")
								ControlRef control;
		
								GetControlByID(window, &kVerticalScrollBar, &control);
								SetControl32BitValue(control, DEFAULT_VERTICAL_SCROLL_BAR_LOC);

								if(pContextInfo && pContextInfo->ptrGP){
									pContextInfo->ptrGP->zoom_bar(DEFAULT_VERTICAL_SCROLL_BAR_LOC);
								}
							}
							
							break;
						case 'open':
							// Open a new file
							if(OpenFiles(window, pContextInfo) == true){
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'plot':
							// Open a new file
							if(OpenFilesGenomePlot(window, pContextInfo) == true){
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'pref':
							// Open the preferences/options window
							if(option_window){
								UpdateOptionsDialog(pContextInfo);
								ShowWindow(option_window);
								SelectWindow(option_window);
							}
							break;
						case 'momt':
							// Toggle momentum 
							if(pContextInfo && pContextInfo->ptrGP){
								pContextInfo->ptrGP->spin_momentum( !pContextInfo->ptrGP->spin_momentum() );
								
								// Update the check box for all centric menu items
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex,
								   pContextInfo->ptrGP->spin_momentum());
								   
								//GetWindowPortBounds (window, &rectPort);
								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'vsze':
							// Set the view to be annotation size-centric
							if(pContextInfo && pContextInfo->ptrGP){
								pContextInfo->ptrGP->mode(DNAMol::SIZE_CENTRIC);
								
								// Update the check box for all centric menu items
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex,
								   true);
								
								// This is "super-kludgey". The menus are in order under the view menu
								// (with "size..." above "gene...")
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex + 1,
								   false);
								
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'vgen':
							// Set the view to be gene-centric
							if(pContextInfo && pContextInfo->ptrGP){
								pContextInfo->ptrGP->mode(DNAMol::GENE_CENTRIC);
								
								// Update the check box
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex,
								   true);
								
								// This is "super-kludgey". The menus are in order under the view menu
								// (with "size..." above "gene...")
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex - 1,
								   false);

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'magn': // Magnify
							if(pContextInfo && pContextInfo->ptrGP){
								
								pContextInfo->ptrGP->magnify( !pContextInfo->ptrGP->magnify() );
								
								// Update the check box
								CheckMenuItem (
								   command.menu.menuRef,
								   command.menu.menuItemIndex,
								   pContextInfo->ptrGP->magnify() );
								   
								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'redf': // Restore the default view
							if(pContextInfo && pContextInfo->ptrGP){
							
								pContextInfo->ptrGP->fit_to_window();
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'exps':
							ExportPS(window, pContextInfo);                                                    
							break;
						case 'expt':
							ExportGenome(window, pContextInfo);
							break;
						case 'exgf':
							ExportGIF(window, pContextInfo);                                                    
							break;
						case 'expg':
							// Open the Exportwindow
							if(export_window){
								ShowWindow(export_window);
								SelectWindow(export_window);
							}                                                  
							break;
						case 'down': // Download ASN data from NCBI
							// Open the Download window
							if(download_window){
								ShowWindow(download_window);
								SelectWindow(download_window);
							}
							break;
						case 'dwns': // Search button on the download dialog
							if(pContextInfo){
								DownloadSearch(pContextInfo->ptrGP);
							}
							break;
						case 'dwnl': // Load button on the download dialog
							if(pContextInfo){
								if(DownloadLoad(pContextInfo->ptrGP, false) == true){
									InvalWindowRect(window, &rectPort);
								}                                                    
							}
							break;
						case 'dwsv': // Load & save button on the download dialog
							if(pContextInfo){
								if(DownloadLoad(pContextInfo->ptrGP, true) == true){
									InvalWindowRect(window, &rectPort);
								}                                                    
							}
							break;
						case 'delg':
							// Delete selected genomes
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->delete_selection()){
								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'iman':
							
							if(pContextInfo){
								// Import annotations from a file
								if(ImportAnnotation(window, pContextInfo) == true){
									InvalWindowRect(window, &rectPort);
								}
							}
							break;
						case 'edan':
							// Open the annotaiton edit dialog
							if(annotation_window){
							
								if(pContextInfo && pContextInfo->ptrGP){
									if(UpdateAnnotationDialog(pContextInfo->ptrGP) == true){
								
										ShowWindow(annotation_window);
										SelectWindow(annotation_window);
									}
								}
							}
							
							break;
						case 'anap':
							// Apply the settings from the annotaiton edit dialog
							if(annotation_window){
							
								if(pContextInfo && pContextInfo->ptrGP){
									if( ApplyAnnotation(pContextInfo->ptrGP) == true ){
									
										InvalWindowRect(window, &rectPort);
									}
								}
							}

							break;
						case 'ancl':
							// Close the annotaiton edit dialog
							if(annotation_window){
								HideWindow(annotation_window);
							}

							break;
						case 'dela':
							// Delete selected annotations
							if(pContextInfo && pContextInfo->ptrGP){
								if(pContextInfo->ptrGP->delete_annotation() == true){
									InvalWindowRect(window, &rectPort);
								}
								else{
									reportError("Please select at least one annotation to delete");
								}
							}
							
							break;
						case 'sall':
							// Select all genomes
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->select_all()){
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'pkpr':
							// Open the Primer dialog for primer picking
							if(primer_window){
								ShowWindow(primer_window);
								SelectWindow(primer_window);
							}
							break;
						case 'prok':
							// Pick primers
							if(pContextInfo && pContextInfo->ptrGP){
							
								if(pContextInfo->ptrGP->num_select() == 0){
									reportError("Please select at least one genome");
									break;
								}
								
								if(PickPrimers(pContextInfo->ptrGP) == true){
									InvalWindowRect(window, &rectPort);
								}
							}
							
							break;
						case 'prcl':
							// Close the primer window
							if(primer_window){
								HideWindow(primer_window);
							}
							break;
						case 'dlpr':
							// Delete all primers for the selected genome
							if(pContextInfo && pContextInfo->ptrGP){
								if(pContextInfo->ptrGP->num_select() == 0){
									reportError("Please select at least one genome");
									break;
								}
								
								pContextInfo->ptrGP->clear_primers();
								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'find':
							// Open the find window
							if(find_window){
								ShowWindow(find_window);
								SelectWindow(find_window);
							}
							break;
						case 'hpcr':
							// Open the hybridization window
							if(hybrid_window){
								ShowWindow(hybrid_window);
								SelectWindow(hybrid_window);
							}
							break;
						case 'hmol':
							// Open the MOLigo window
							if(moligo_window){
								ShowWindow(moligo_window);
								SelectWindow(moligo_window);
							}
							break;
						case 'fagn': // menu: find again
						case 'fnxt': // find window: find again
						case 'hnxt': // hybridize window show next
						case 'mnxt': // MOLigo window show next
							// We need to get a valid Genome window here!
							if(pContextInfo && pContextInfo->ptrGP){
								pContextInfo->ptrGP->search_next();
								
								InvalWindowRect(window, &rectPort);
							}
							break;
						
						case 'fant':
						case 'hant':
						case 'mant':
							// Annotate the selected sequence(s)
							// We need to get a valid Genome window here!
							if(pContextInfo && pContextInfo->ptrGP){
							
								bool annot_ret = pContextInfo->ptrGP->annotate_selection();
								
								if(annot_ret == false){
									reportError ("There must be at lest one selected sequence "
										"region to annotate");
								}
								else{
									InvalWindowRect(window, &rectPort);
								}
							}
							break;
						case 'ffnd': // find window: find
							// Extract the search key and string from the find window
							// and perform the search. Note that find_window is a global variable
							// and does not need to be passed to the FinSeq function.
							if(pContextInfo && pContextInfo->ptrGP && FindSeq(pContextInfo->ptrGP)){
							
								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'hprm': // hybridize window: primer search
							if(pContextInfo && HybridizePrimers(pContextInfo->ptrGP)){

								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'hprb': // hybridize window: probe search
							if(pContextInfo && pContextInfo->ptrGP && HybridizeProbe(pContextInfo->ptrGP)){

								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'mprm': // MOLigo window: primer search
							if(pContextInfo && pContextInfo->ptrGP && MOLigoPrimers(pContextInfo->ptrGP)){

								InvalWindowRect(window, &rectPort);
							}
							
							break;
						case 'exfm': // Export window format change
							UpdateExportWindow();
							break;
						case 'fclo': // close the find window
							if(find_window){
								HideWindow(find_window);
							}
							break;
						case 'hclo': // close the hybridize window
							if(hybrid_window){
								HideWindow(hybrid_window);
							}
							break;
						case 'mclo': // close the MOLigo window
							if(moligo_window){
								HideWindow(moligo_window);
							}
							break;
						case 'dclo': // close the download window
							if(download_window){
								HideWindow(download_window);
							}
							break;
						case 'iclo': // Close the info window
							if(info_window){
								HideWindow(info_window);
							}                                                    
							break;
						case 'oclo': // Close the options window
							if(option_window){
								HideWindow(option_window);
							}                                                    
							break;
						case 'excl': // close the export window
							if(export_window){
								HideWindow(export_window);
							}
							break;
						case 'hide':
							// Hide a genome
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->hide_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'uhid':
							// Un-hide a genome
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->show_all()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'hdan':
							// Hide genome annotation
							if(pContextInfo && pContextInfo->ptrGP && 
								pContextInfo->ptrGP->hide_selection_annotation()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'uhan':
							// Un-hide genome annotation
							if( pContextInfo && pContextInfo->ptrGP && 
								pContextInfo->ptrGP->show_annotation() ){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'gc  ':
							// Display % GC
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_gc_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'at  ':
							// Display % AT
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_at_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'gcsk':
							// Display GC Skew
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_gc_skew_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'atsk':
							// Display AT Skew
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_at_skew_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'keto':
							// Display Keto Skew
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_keto_skew_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'purn':
							// Display Purine Skew
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_purine_skew_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'tm  ':
							// Display DNA melting
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_tm_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'cons':
							// Display DNA conservation score
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_conservation_plot()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'pl  ':
							// Display a user plot
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->toggle_genome_plot_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'mup ':
							// Move selected genomes up
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->shift_selection_up()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'mdwn':
							// Move selected genomes down
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->shift_selection_down()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'more':
							// Show one more gnenome on screen
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->key_plus()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'less':
							// Show one less genome on screen
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->key_minus()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'comp':
							// Take the complement of selected genomes
							if(pContextInfo && pContextInfo->ptrGP && pContextInfo->ptrGP->complement_selection()){

								InvalWindowRect(window, &rectPort);
							}
							break;
						case 'dna ':
							// Show DNA sequence for the selected annotations
							// (FASTA for now).
							if(pContextInfo && pContextInfo->ptrGP){
								std::string info = pContextInfo->ptrGP->selected_seq_as_dna();
								
								if(info != ""){
									DisplayInfo(info);
								}
							}
							
							break;
						case 'prot':
							// Show amino acid sequence for the selected annotations
							// (FASTA for now).
							if(pContextInfo && pContextInfo->ptrGP){
								std::string info = pContextInfo->ptrGP->selected_seq_as_protein();
								
								if(info != ""){
									DisplayInfo(info);
								}
							}
							break;
						case 'cbak':
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_BACKGROUND_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_BACKGROUND_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_BACKGROUND_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->background_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}
							break;
						case 'csel':
							// Set the selection color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_SELECT_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_SELECT_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_SELECT_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->select_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'ctxt':
							// Set the text color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_TEXT_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_TEXT_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_TEXT_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->text_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'cgne':
							// Set the gene color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_GENE_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_GENE_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_GENE_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->gene_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'cpgn':
							// Set the pseudo-gene color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_PSEUDO_GENE_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_PSEUDO_GENE_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_PSEUDO_GENE_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->pseudo_gene_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'ccds':
							// Set the CDS (i.e. orf) color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_CDS_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_CDS_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_CDS_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->cds_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'cann':
							// Set the annotation color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_ANNOT_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_ANNOT_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_ANNOT_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->annot_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'crna':
							// Set the RNA color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_RNA_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_RNA_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_RNA_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->rna_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'cint':
							// Set the intergenic space color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_INTERGENIC_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_INTERGENIC_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_INTERGENIC_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->intergenic_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						case 'ccst':
							// Set the custom annotation color
							if(pContextInfo && pContextInfo->ptrGP){
								NColorPickerInfo colorInfo;
								
								memset(&colorInfo, 0, sizeof(colorInfo));
								
								// 16 bits per channel (48 bit color) (i.e. [0 ... 65535])
								colorInfo.theColor.color.rgb.red = (UInt16)(GV_DEFAULT_CUSTOM_R*color_8_to_16);
								colorInfo.theColor.color.rgb.green = (UInt16)(GV_DEFAULT_CUSTOM_G*color_8_to_16);
								colorInfo.theColor.color.rgb.blue = (UInt16)(GV_DEFAULT_CUSTOM_B*color_8_to_16);
								
								result = NPickColor(&colorInfo);
								
								if((result == noErr) && (colorInfo.newColorChosen)){
								
									pContextInfo->ptrGP->custom_color(
										(unsigned char)(colorInfo.theColor.color.rgb.red*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.green*color_16_to_8),
										(unsigned char)(colorInfo.theColor.color.rgb.blue*color_16_to_8));
									
									// Update the background color
									InvalWindowRect(window, &rectPort);
								}
							}

							break;
						}
					break;
				case kEventCommandUpdateStatus: // menu updates
					GetEventParameter (event, kEventParamDirectObject, kEventParamHICommand, NULL, sizeof(command), NULL, &command); // get command
					switch (command.commandID) {
						case kHICommandClose:
							if (pContextInfo)
								EnableMenuItem (GetMenuHandle (kMainMenu), kCloseMenuItem);
							else 
								DisableMenuItem (GetMenuHandle (kMainMenu), kCloseMenuItem);
							break;
						case 'gogo':
							if (pContextInfo) {
								EnableMenuItem (GetMenuHandle (kMainMenu), kAnimateMenuItem);
							} else { 
								DisableMenuItem (GetMenuHandle (kMainMenu), kAnimateMenuItem);
								CheckMenuItem (GetMenuHandle (kMainMenu), kAnimateMenuItem, kAnimateState);
							}
							break;
						case 'info':
							if (pContextInfo) {
								EnableMenuItem (GetMenuHandle (kMainMenu), kInfoMenuItem);
							} else { 
								DisableMenuItem (GetMenuHandle (kMainMenu), kInfoMenuItem);
								CheckMenuItem (GetMenuHandle (kMainMenu), kInfoMenuItem, kInfoState);
							}
							break;
					}
					result = noErr;
					break;
			}
			break;
	}
	
	}
	catch(const char *error){
		ErrorDialog(error);
	}
	catch(...){
		ErrorDialog("An unhandled error has occured", "Please report this error to jgans@lanl.gov");
	}
	
    return result;
}

#pragma mark ==== main ====

int main(int argc, char* argv[])
{
    OSStatus		err;
    EventHandlerRef	ref;
    EventTypeSpec	list[] = {{ kEventClassCommand,  kEventProcessCommand },
							{ kEventClassCommand,  kEventCommandUpdateStatus },
							{ kEventClassMouse, kEventMouseDown },
							{ kEventClassMouse, kEventMouseUp }, 
							{ kEventClassMouse, kEventMouseMoved },
							{ kEventClassMouse, kEventMouseDragged },
							{ kEventClassMouse, kEventMouseWheelMoved } };
    
    ProcessSerialNumber psn = { 0, kCurrentProcess };
    EventRef windowEvent;

    gStartTime = UpTime (); // get app start time
	
    // Create a Nib reference passing the name of the nib file (without the .nib extension)
    // CreateNibReference only searches into the application bundle.
    err = CreateNibReference(CFSTR("main"), &nibRef);
    require_noerr( err, CantGetNibRef );
    
    err = SetMenuBarFromNib(nibRef, CFSTR("MainMenu"));
    require_noerr( err, CantSetMenuBar );
	
    gEvtHandler = NewEventHandlerUPP(appEvtHndlr);
    err = InstallApplicationEventHandler (gEvtHandler, GetEventTypeCount (list), list, 0, &ref);
    
    gWinEvtHandler = NewEventHandlerUPP(windowEvtHndlr); 

	// Allow drag-and-drop opening of files
	AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, NewAEEventHandlerUPP(ReceiveDroppedFiles), 0, false);
	
    getCurrentCaps ();
    
    // Initialize the text control
    err = TXNInitTextension(NULL, 0, kTXNUseTemporaryMemoryMask);
    
    if(err != noErr){
        throw __FILE__ ":main: Failure in TXNInitTextension";
    }
    
    // Create the find window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Find"), &find_window); // build window
    
    // Add a scrollable text window to the find dialog
    if(AddTextViewWidget(find_window, kFindSearchResultsView, kFindSearchResults, true) != noErr){
        throw __FILE__ ":main: Unable to attach text view to find dialog";
    }
    
	// Create the hybridizie window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Hybridize"), &hybrid_window); // build window
    
    // Add a scrollable text window to the hybridizie dialog
    if(AddTextViewWidget(hybrid_window, kHybridResultsView, kHybridResultsEdit, true) != noErr){
        throw __FILE__ ":main: Unable to attach text view to hybridize dialog";
    }
	
	// Create the MOLigo window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("MOLIGO"), &moligo_window); // build window
    
    // Add a scrollable text window to the hybridizie dialog
    if(AddTextViewWidget(moligo_window, kMoligoResultsView, kMoligoResultsEdit, true) != noErr){
        throw __FILE__ ":main: Unable to attach text view to Moligo dialog";
    }
	
    // Create the info window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Info"), &info_window); // build window
    
    // Add a scrollable text window to the info dialog
    if(AddTextViewWidget(info_window, kInfoView, kInfoText, true) != noErr){
        throw __FILE__ ":main: Unable to attach text view to info dialog";
    }
    
	// Create the primer window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Primer"), &primer_window); // build window
	
    // Create the info window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Download"), &download_window); // build window
    
    ConfigureDataBrowser(download_window, kDownloadSearchResults);
    
	// Create the Options window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Options"), &option_window); // build window
    
	ControlRef control;

	GetControlByID(option_window, &kOptionsFont, &control);
	SetControlAction(control, fProc);
	
	GetControlByID( option_window, &kOptionsGC, &control );
	SetControlAction(control, fProc);
	
	GetControlByID( option_window, &kOptionsTm, &control );
	SetControlAction(control, fProc);
	
	GetControlByID( option_window, &kOptionsTmSalt, &control );
	SetControlAction(control, fProc);
	
	GetControlByID( option_window, &kOptionsTmStrand, &control );
	SetControlAction(control, fProc);
	
	// Create the export window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Export"), &export_window); // build window

	// Create the export window (we want a global window that will have persistance - i.e.
    // can be opened and closed without losing settings).
    CreateWindowFromNib (nibRef, CFSTR("Annotate"), &annotation_window); // build window
	
	// Add a scrollable text window to the annotation edit dialog
    if(AddTextViewWidget(annotation_window, kAnnotNoteView, kAnnotNote, false) != noErr){
        throw __FILE__ ":main: Unable to attach text view to annotation dialog";
    }
	
	// Setup the mouse cursor
    InitCursor ();
    
    // ensure we know when display configs are changed
    gConfigEDMUPP = NewDMExtendedNotificationUPP (handleConfigDMEvent); // for display change notification
    DMRegisterExtendedNotifyProc (gConfigEDMUPP, NULL, NULL, &psn);
    
    if(err == noErr) { // post new window event
        err = MacCreateEvent (nil, kEventClassCommand, kEventProcessCommand, 0, kEventAttributeNone, &windowEvent);
		
        if(err == noErr) {
            HICommand command; // set HI command parameter...
            command.commandID = kHICommandNew; command.attributes = 0; 
            command.menu.menuRef = 0; command.menu.menuItemIndex = 0;
            err = SetEventParameter(windowEvent, kEventParamDirectObject, kEventParamHICommand, sizeof(command), &command);
            
            if(err == noErr){
                err = PostEventToQueue (GetCurrentEventQueue(), windowEvent, kEventPriorityHigh);
            }
            
            ReleaseEvent (windowEvent);
        }
    }

    // Call the event loop
    RunApplicationEventLoop();
	
    // Exiting...

    if (gConfigEDMUPP) { // dispose UPP for DM notifications
        DisposeDMExtendedNotificationUPP (gConfigEDMUPP);
        gConfigEDMUPP = NULL;
    }

    if (gDisplayCaps) { // dispose diaplsy capabilities info
        free (gDisplayCaps);
        gDisplayCaps = NULL;
    }
    
    if(find_window){
        DisposeWindow(find_window);
    }
    
	if(hybrid_window){
        DisposeWindow(hybrid_window);
    }
	
	if(moligo_window){
        DisposeWindow(moligo_window);
    }
	
	if(download_window){
        DisposeWindow(download_window);
    }
	
	if(info_window){
        DisposeWindow(info_window);
    }

	if(option_window){
        DisposeWindow(option_window);
    }
	
	if(primer_window){
		DisposeWindow(primer_window);
	}
	
	if(export_window){
		DisposeWindow(export_window);
	}
	
	if(annotation_window){

		DisposeWindow(annotation_window);
	}
	
//CantCreateWindow:
CantSetMenuBar:
CantGetNibRef:

    // We don't need the nib reference anymore.
    DisposeNibReference (nibRef);

	return err;
}

