// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#ifndef __TEXT_FONT_GL
#define __TEXT_FONT_GL

#ifdef _MAC_OSX
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

#include <string>

#define		GL_CHAR_START	(int)(' ')
#define		GL_CHAR_STOP	(int)(126)
#define		GL_CHAR_COUNT	(GL_CHAR_STOP - GL_CHAR_START + 1)

// A class for handling OpenGL font metric information
// Currently, we only store the text DX size
class FontMetricGL{
private:
	GLfloat dx[GL_CHAR_COUNT];
public:
	FontMetricGL()
	{
		for(unsigned int i = 0;i < GL_CHAR_COUNT;i++){
			dx[i] = 1.0f;
		}
	};

	inline GLfloat& operator[](const unsigned char m_i)
	{
		#ifdef _DEBUG
		if(m_i < GL_CHAR_START){
			throw "FontMetricGL: char underflow";
		}

		if(m_i > GL_CHAR_STOP){
			throw "FontMetricGL: char overflow";
		}

		#endif // _DEBUG

		return dx[m_i - GL_CHAR_START];
	};

	inline GLfloat size(const std::string &m_str) const
	{
		unsigned int str_len = m_str.size();
		GLfloat str_size = 0.0f;

		for(unsigned int i = 0;i < str_len;i++){
			str_size += dx[ m_str[i] - GL_CHAR_START ];
		}

		return str_size;
	};
};


#endif // __TEXT_FONT_GL

