// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include <fstream>

using namespace std;

void GenomePage::export_fasta_genome(const std::string &m_filename)
{
	// Make sure that we have a valid file name
	ofstream fout( m_filename.c_str() );
	
	if(!fout){
		throw "Unable to open file for Fasta export";
	}
	
	list<Genome>::iterator iter;
	
	// If no genomes are selected, export ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
				iter->export_fasta_genome(fout);
			}
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
				iter->export_fasta_genome(fout);
			}
		}
	}
	
	fout.close();
}

void GenomePage::export_fasta_dna(const std::string &m_filename)
{
	// Make sure that we have a valid file name
	ofstream fout( m_filename.c_str() );
	
	if(!fout){
		throw "Unable to open file for Fasta export";
	}
	
	list<Genome>::iterator iter;
	
	// If no genomes are selected, export ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
				iter->export_fasta_dna(fout, annotations_to_export);
			}
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
				iter->export_fasta_dna(fout, annotations_to_export);
			}
		}
	}
	
	fout.close();
}

void GenomePage::export_fasta_protein(const std::string &m_filename)
{
	// Make sure that we have a valid file name
	ofstream fout( m_filename.c_str() );
	
	if(!fout){
		throw "Unable to open file for Fasta export";
	}
	
	list<Genome>::iterator iter;
	
	// If no genomes are selected, export ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
				iter->export_fasta_protein(fout, annotations_to_export);
			}
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
				iter->export_fasta_protein(fout, annotations_to_export);
			}
		}
	}
	
	fout.close();
}

void GenomePage::export_gbk(const std::string &m_filename)
{
	// Make sure that we have a valid file name
	ofstream fout( m_filename.c_str() );
	
	if(!fout){
		throw "Unable to open file for GBK export";
	}
	
	list<Genome>::iterator iter;
	
	// If no genomes are selected, export ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
				iter->export_gbk(fout, annotations_to_export);
			}
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
				iter->export_gbk(fout, annotations_to_export);
			}
		}
	}
	
	fout.close();
}

void GenomePage::export_ptt(const std::string &m_filename)
{
	// Make sure that we have a valid file name
	ofstream fout( m_filename.c_str() );
	
	if(!fout){
		throw "Unable to open file for PTT export";
	}
	
	list<Genome>::iterator iter;
	
	// If no genomes are selected, export ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
			
				// Only export a single genome
				iter->export_ptt(fout, annotations_to_export);
				break;
			}
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
			
				// Only export a single genome
				iter->export_ptt(fout, annotations_to_export);
				break;
			}
		}
	}
	
	fout.close();
}

void Genome::export_fasta_genome(ofstream &m_fout)
{
	mol.export_fasta_genome(m_fout);
}

void Genome::export_fasta_dna(ofstream &m_fout, const list<unsigned int> &m_annot)
{
	mol.export_fasta_dna(m_fout, m_annot);
}

void Genome::export_fasta_protein(ofstream &m_fout, const list<unsigned int> &m_annot)
{
	mol.export_fasta_protein(m_fout, m_annot);
}

void Genome::export_gbk(ofstream &m_fout, const list<unsigned int> &m_annot)
{
	mol.export_gbk(m_fout, m_annot);
}

void Genome::export_ptt(ofstream &m_fout, const list<unsigned int> &m_annot)
{
	mol.export_ptt(m_fout, m_annot);
}
