// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include "univout.h"

using namespace std;

bool GenomePage::l_button_down(const int &m_x, const int &m_y)
{	
	mouse_status |= GP_MOUSE_L_BUTTON_DOWN;
	mouse_l_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_l_start[Y] = height - m_y;
	#else
	mouse_l_start[Y] = m_y;
	#endif	
	
	// Stop spinning, the use is trying to move something
	spinning = false;
	spin_dx = 0.0f;

	// Deselect any previously selected genome(s). Don't clear the 
	// translation buffer
	clear_selection(false);
	
	// Pick a Genome
	list<Genome>::iterator pick_iter;
	
	if(pick(mouse_l_start[X], mouse_l_start[Y], pick_iter)){
		pick_iter->select(true);
	}

	return true;
}

bool GenomePage::l_button_ctrl_down(const int &m_x, const int &m_y)
{	
	mouse_status |= GP_MOUSE_L_BUTTON_DOWN;
	mouse_l_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_l_start[Y] = height - m_y;
	#else
	mouse_l_start[Y] = m_y;
	#endif	
	
	// Stop spinning, the use is trying to select something
	spinning = false;
	spin_dx = 0.0f;

	// Don't clear any existing selections

	// Pick a Genome
	list<Genome>::iterator pick_iter;
	
	int pick_state = pick(mouse_l_start[X], mouse_l_start[Y], pick_iter);

	if(pick_state == PICKED_GENOME){
		// Toggle the selection state of the *genome* that the user has selected.
		// Do not toggle if the user picks an annotation object
		pick_iter->select( !pick_iter->select() );
	} 
	else if(pick_state == PICKED_ANNOT){
		// If a user selects an annotation, the selection state of the genome is
		// always true (i.e. selected)
		pick_iter->select(true);
	}

	return true;
}

bool GenomePage::l_button_shift_down(const int &m_x, const int &m_y, string &m_seq)
{
	mouse_status |= GP_MOUSE_L_BUTTON_DOWN;
	mouse_l_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_l_start[Y] = height - m_y;
	#else
	mouse_l_start[Y] = m_y;
	#endif	
	
	// Stop spinning, the use is trying to select something
	spinning = false;
	spin_dx = 0.0f;

	// Deselect any previously selected genome(s). Don't clear the translation buffer!
	clear_selection(false);

	// Pick a Genome
	list<Genome>::iterator pick_iter;
	
	if(pick(mouse_l_start[X], mouse_l_start[Y], pick_iter)){
		pick_iter->select(true);

		// Add the selected base to the translation buffer.
		// The translation buffer can have either 0 or 1 elements
		if( translation_buffer_base.empty() ){
			// The buffer is currently empty, fill it with the selected base
			// and genome.
			translation_buffer_base.push_back(cur_base_loc());
			translation_buffer_genome.push_back(pick_iter);

			uout << "Added first base to translation buffer" << uout_endl;
		}
		else{
			// The translation buffer currently has one element in it -- are
			// the genomes the same?
			if( pick_iter == translation_buffer_genome.back() ){
				// Yes -- we have a match!

				// First, show the DNA
				m_seq = pick_iter->sequence(translation_buffer_base.back(), cur_base_loc());

				// Then show the Protein
				m_seq += pick_iter->translate_sequence(translation_buffer_base.back(), cur_base_loc());
			}
			else{
				uout << "Genomes didn't match, translation canceled" << uout_endl;
			}

			// Clear the buffer for the next use
			translation_buffer_base.clear();
			translation_buffer_genome.clear();

			uout << "Translation buffer cleared" << uout_endl;
		}
	}

	return true;
}

bool GenomePage::l_button_dbl_down(const int &m_x, const int &m_y, string &m_info)
{
	// The left mouse button is NOT down!
	mouse_status &= ~GP_MOUSE_L_BUTTON_DOWN;

	mouse_l_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_l_start[Y] = height - m_y;
	#else
	mouse_l_start[Y] = m_y;
	#endif	
	
	// Stop spinning, the use is trying to select something
	spinning = false;
	spin_dx = 0.0f;

	// Deselect any previously selected genome(s). Don't clear
	// the translation buffer
	clear_selection(false);
	
	// Pick a Genome
	list<Genome>::iterator pick_iter;
	
	if(pick(mouse_l_start[X], mouse_l_start[Y], pick_iter)){
		pick_iter->select(true);

		// Is there any info to display?
		m_info = pick_iter->select_info();
	}

	return true;
}

bool GenomePage::l_button_up(const int &m_x, const int &m_y)
{
	mouse_status &= ~GP_MOUSE_L_BUTTON_DOWN;

	if(momentum && (spin_dx != 0.0f) ){
		// Start spinning
		spinning = true;
		last_elapsed_time = 0L;

		return true;
	}

	return false;
}

bool GenomePage::m_button_down(const int &m_x, const int &m_y)
{
	mouse_status |= GP_MOUSE_M_BUTTON_DOWN;
	mouse_m_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_m_start[Y] = height - m_y;
	#else
	mouse_m_start[Y] = m_y;
	#endif	

	// Stop spinning, the use is trying to move something
	spinning = false;

	return false;
}

bool GenomePage::m_button_up(const int &m_x, const int &m_y)
{
	mouse_status &= ~GP_MOUSE_M_BUTTON_DOWN;

	return false;
}

// The right mouse button controls the zoom level
bool GenomePage::r_button_down(const int &m_x, const int &m_y)
{
	mouse_status |= GP_MOUSE_R_BUTTON_DOWN;
	mouse_r_start[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_r_start[Y] = height - m_y;
	#else
	mouse_r_start[Y] = m_y;
	#endif	

	return false;
}

bool GenomePage::r_button_up(const int &m_x, const int &m_y)
{
	mouse_status &= ~GP_MOUSE_R_BUTTON_DOWN;

	return false;
}

bool GenomePage::mouse_wheel(const int &m_x, const int &m_y, const int &m_delta)
{
	// This function should never be called with m_delta == 0, but better
	// safe than sorry.
	if(m_delta == 0){
		return false;
	}

	// The mouse wheel zooms by zoom_fract
	const float zoom_fract = 0.20f;

	if(m_delta > 0){
		zoom((int)(base_window*(1.0f + zoom_fract)));
	}
	else{
		zoom((int)(base_window*(1.0f - zoom_fract)));
	}

	return true;
}

bool GenomePage::mouse_move(const int &m_x, const int &m_y)
{
	// Do we need to update the mouse guide or the translation guide (test before we set the 
	// new mouse location)?
	bool status = ( (zoom_level == ZOOM_LEVEL_SEQ) || (translation_buffer_base.empty() == false) ) 
		&& base_loc_change(m_x);

	mouse_loc[X] = m_x;

	#ifdef INVERT_MOUSE_Y
	mouse_loc[Y] = height - m_y;
	#else
	mouse_loc[Y] = m_y;
	#endif	

	if(mouse_status & GP_MOUSE_L_BUTTON_DOWN){
	
		// Has there been any movement?
		if((mouse_loc[X] == mouse_l_start[X]) && (mouse_loc[Y] == mouse_l_start[Y])){
			return false;
		}

		// Shift the selected genome(s)
		if(mouse_loc[X] != mouse_l_start[X]){
		
			// Scale dx by 2.0 since the shift function moves in fractions of
			// half the window size. Not intuitive. My bad.
			spin_dx = (2.0f*(mouse_loc[X] - mouse_l_start[X]))/width;

			list<Genome>::iterator iter;
			bool ret_value = false;

			for(iter = genome_list.begin();iter != genome_list.end();iter++){
				if(iter->select()){
					ret_value = true;

					iter->shift(spin_dx, true /* Accumulate to the shift buffer */);
				}
			}
		}

		// Using the left mouse button to both shift and scroll induces vertigo!
		// We're scrolling the view up or down
		//if(mouse_loc[Y] != mouse_l_start[Y]){
		//	GLfloat dy = ((GLfloat)(mouse_loc[Y]) - mouse_l_start[Y])/height;

			// The factor of 2 is a fudge factor to scroll faster
		//	scroll(-2.0f*dy);
		//}

		// Update the shift location
		mouse_l_start[X] = mouse_loc[X];
		mouse_l_start[Y] = mouse_loc[Y];
		
		auto_pick(mouse_loc[X], mouse_loc[Y]);
		
		return true;
	}

	if(mouse_status & GP_MOUSE_M_BUTTON_DOWN){
		
		// Has there been any movement?
		if((mouse_loc[X] == mouse_m_start[X]) && (mouse_loc[Y] == mouse_m_start[Y])){
			return false;
		}

		// We're scrolling the *entire* view up/down and left/right
		const GLfloat dy = ((GLfloat)(mouse_loc[Y]) - mouse_m_start[Y])/height;
		const GLfloat dx = ((GLfloat)(mouse_loc[X]) - mouse_m_start[X])/width;

		// The factor of 2 is a fudge factor to scroll faster
		scroll(-2.0f*dy);

		list<Genome>::iterator iter;
		

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			iter->shift(dx, true /* Accumulate to the shift buffer */);
		}

		// Update the scroll location
		mouse_m_start[X] = mouse_loc[X];
		mouse_m_start[Y] = mouse_loc[Y];
		
		auto_pick(mouse_loc[X], mouse_loc[Y]);
		
		return true;
	}

	if(mouse_status & GP_MOUSE_R_BUTTON_DOWN){
		
		// Has there been any movement?
		if((mouse_loc[X] == mouse_r_start[X]) && (mouse_loc[Y] == mouse_r_start[Y])){
			return false;
		}

		// We're adjusting the zoom level
		// +x --> zoom in
		// -x --> zoom out
		GLfloat dx = ((GLfloat)(mouse_loc[X]) - mouse_r_start[X])/width;

		// Zoom by fractional amount of mouse movement
		if(dx > 0.0f){
			zoom((int)ceil(base_window*(1.0 + dx)));
		}
		else{
			zoom((int)floor(base_window*(1.0 + dx)));
		}

		// Update the zoom location
		mouse_r_start[X] = mouse_loc[X];
		mouse_r_start[Y] = mouse_loc[Y];
		
		auto_pick(mouse_loc[X], mouse_loc[Y]);
		
		return true;
	}
	
	return status || auto_pick(mouse_loc[X], mouse_loc[Y]);
}

bool GenomePage::scroll_bar(const int &m_loc, const int &m_width)
{
	// The scroll bar overrides any mouse-induced spinning
	spinning = false;

	// Initialize the last scroll bar position (if needed)
	if(scroll_loc < 0){
		scroll_loc = m_width/2;
	}
	
	// Shift by m_delta
	const float dx = -( 2.0f*(m_loc - scroll_loc) )/m_width;
	list<Genome>::iterator iter;

	// Move everything, regardless of which genomes are selected
	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->shift(dx, true);
	}

	// Update the new scroll bar position
	scroll_loc = m_loc;
	
	return true;
}

bool GenomePage::zoom_bar(const int &m_loc, const int &m_width)
{
	// Initialize the last zoom bar position (if needed)
	if(zoom_loc < 0){
		zoom_loc = m_width/2;
	}
	
	// Zoom by dx. The sign of dx is chosen to be compatible with the 
	// scroll wheel
	const float dx = -2.0*(float)( (m_loc - zoom_loc) )/m_width;
	
	// Zoom by fractional amount of mouse movement
	if(dx > 0.0f){
		zoom((int)ceil(base_window*(1.0 + dx)));
	}
	else{
		zoom((int)floor(base_window*(1.0 + dx)));
	}

	// Update the new zoom bar position
	zoom_loc = m_loc;
	
	return true;
}
